/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Map;

class FileUploader
{
	private OutputStream m_outStream;
	private Map<String, String> m_headers;

	private static boolean m_bDisabled = false;

	FileUploader(OutputStream outStream, Map<String, String> headers)
	{
		m_outStream = outStream;
		m_headers = headers;
	}

	public static void setDisabled(boolean bDisabled)
	{
		m_bDisabled = bDisabled;
	}

	public void uploadFile(File thisFile, boolean bCheckFileOpen, String contentDisposition) throws Exception
	{
		if (m_bDisabled)
		{
			StringBuffer header = new StringBuffer();
			header.append("HTTP/1.0 404 Not Found\n");
			setHeader(header, "Content-type", "text/html");
			header.append("\n");
			header.append("<html><head><title>Not Found</title></head><body>Not Found</body></html>");
			m_outStream.write(header.toString().getBytes("UTF-8"));
		}
		else if ( bCheckFileOpen && isFileOpen(thisFile) )
		{
			FileInputStream fi = new FileInputStream(thisFile);

			int read = 0;
			byte[] bytes = new byte[DEFAULT_BUFFER_SIZE];
			read = fi.read(bytes);

			String name = Thread.currentThread().getName();
			long prevTime = System.nanoTime();
			long currentTime;
			long prevSize = 0;
			long currentSize = 0;

			StringBuffer header = new StringBuffer();
			header.append("HTTP/1.0 200 OK\n");
			header.append("Content-Disposition: " + contentDisposition + "; filename=\"" + thisFile.getName() + "\"\n");
			header.append("\n");
			m_outStream.write(header.toString().getBytes("UTF-8"));

			while (true)
			{
				while (read > -1)
				{
					m_outStream.write(bytes, 0, read);

					currentSize += read;
					currentTime = System.nanoTime();
					if (currentTime > (prevTime + 5 * 1000000000L))
					{
						setStatus(isFileOpen(thisFile), name, prevSize, currentSize, thisFile.length(), prevTime, currentTime);
						prevSize = currentSize;
						prevTime = currentTime;
					}

					read = fi.read(bytes);
					//Thread.sleep(1000);	// Limit speed for testing.
				}
				if ( isFileOpen(thisFile) )
				{
					Thread.sleep(1000);
					read = fi.read(bytes);
					currentSize += read;
				}
				else
				{
					read = fi.read(bytes);
					while (read > -1)
					{
						m_outStream.write(bytes, 0, read);

						currentSize += read;
						currentTime = System.nanoTime();
						if (currentTime > (prevTime + 5 * 1000000000L))
						{
							setStatus(isFileOpen(thisFile), name, prevSize, currentSize, thisFile.length(), prevTime, currentTime);
							prevSize = currentSize;
							prevTime = currentTime;
						}

						read = fi.read(bytes);
					}
					break;
				}
			}
			fi.close();
		}
		else
		{
			/*
			FileInputStream fi = new FileInputStream(thisFile);

			int read = 0;
			byte[] bytes = new byte[DEFAULT_BUFFER_SIZE];
			read = fi.read(bytes);

			StringBuffer header = new StringBuffer();
			header.append("HTTP/1.0 200 OK\n");
			header.append("Content-Disposition: " + contentDisposition + "; filename=\"" + thisFile.getName() + "\"\n");
			long fileLength = thisFile.length();
			header.append("Content-Length: " + fileLength + "\n");
			header.append("\n");
			outStream.write(header.toString().getBytes("UTF-8"));

			while (read > -1)
			{
				outStream.write(bytes, 0, read);
				read = fi.read(bytes);
			}
			fi.close();
			 */

			processRequest(thisFile, contentDisposition);
		}
	}

	// Check if file open by another process.
	public static boolean isFileOpen(File thisFile) throws Exception
	{
		if (!thisFile.canWrite())
			return false;

		boolean bFileOpen = false;
		FileOutputStream fo = null;
		try {
			fo = new FileOutputStream(thisFile, true);
		} catch(IOException e) {
			bFileOpen = true;
		} finally {
			if(fo != null)
			{
				try {
					fo.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		return bFileOpen;
	}

	private static void setStatus(StringBuffer buffer, String status)
	{
		buffer.insert(0, status + "\n");
	}

	private static void setHeader(StringBuffer buffer, String header1, String header2)
	{
		buffer.append(header1 + ": " + header2 + "\n");
	}

	private static void setDateHeader(StringBuffer buffer, String header1, long date)
	{
		buffer.append(header1 + ": " + DateUtil.formatDate(new Date(date)) + "\n");
	}

	private long getDateHeader(String header) throws Exception
	{
		String sDate = m_headers.get(header);
		if ( (sDate == null) || (sDate.isEmpty()) )
			return -1;
		return DateUtil.parseDate(sDate).getTime();
	}

	private void writeHeader(StringBuffer header, boolean bInsertLine) throws Exception
	{
		if (bInsertLine)
			header.append("\n");
		m_outStream.write(header.toString().getBytes("UTF-8"));
	}

	private static void setStatus(String name, long prevSize, long currentSize, long totalSize, long prevTime, long currentTime)
	{
		String percent = String.format("%1$,.2f", (double)currentSize / (double)totalSize * 100);
		double speed = (double)(currentSize - prevSize) / ((double)(currentTime - prevTime) / 1000000000L);
		String speedString = String.format("%1$,.2f", speed / 1024);
		int timeRemaining = (int) ((totalSize - currentSize) / speed);
		String timeRemainingString = String.format("%d:%02d:%02d", timeRemaining/3600, (timeRemaining%3600)/60, (timeRemaining%60));
		Thread.currentThread().setName(name + "\n" 
				+ String.format("%,d", currentSize / 1024) + "KB of " + String.format("%,d", totalSize / 1024) + "KB (" + percent + "%)" + "\n"
				+ speedString + "KB/s" + "\n"
				+ timeRemainingString);
	}

	private static void setStatus(boolean bLive, String name, long prevSize, long currentSize, long totalSize, long prevTime, long currentTime)
	{
		if (bLive)
		{
			String speedString = String.format("%1$,.2f", (double)(currentSize - prevSize) / 1024 / ((double)(currentTime - prevTime) / 1000000000L));
			Thread.currentThread().setName(name + "\n" 
					+ String.format("%,d", currentSize / 1024) + "KB" + "\n"
					+ speedString + "KB/s" + "\n"
					+ "Live stream");
		}
		else
			setStatus(name, prevSize, currentSize, totalSize, prevTime, currentTime);
	}

	// Following from http://balusc.blogspot.com/2009/02/fileservlet-supporting-resume-and.html
	private void processRequest(File file, String contentDisposition) throws Exception
	{
		// Validate the requested file ------------------------------------------------------------

		// Prepare some variables. The ETag is an unique identifier of the file.
		String fileName = file.getName();
		long length = file.length();
		long lastModified = file.lastModified();
		String eTag = "\"" + fileName + "_" + length + "_" + lastModified + "\"";
		long expires = System.currentTimeMillis() + DEFAULT_EXPIRE_TIME;
		StringBuffer header = new StringBuffer();


		// Validate request headers for caching ---------------------------------------------------

		// If-None-Match header should contain "*" or ETag. If so, then return 304.
		String ifNoneMatch = m_headers.get("If-None-Match");
		if (ifNoneMatch != null && matches(ifNoneMatch, eTag)) {
			setStatus(header, SC_NOT_MODIFIED);
			setHeader(header, "ETag", eTag); // Required in 304.
			setDateHeader(header, "Expires", expires); // Postpone cache with 1 week.
			writeHeader(header, true);
			return;
		}

		// If-Modified-Since header should be greater than LastModified. If so, then return 304.
		// This header is ignored if any If-None-Match header is specified.
		long ifModifiedSince = getDateHeader("If-Modified-Since");
		if (ifNoneMatch == null && ifModifiedSince != -1 && ifModifiedSince + 1000 > lastModified) {
			setStatus(header, SC_NOT_MODIFIED);
			setHeader(header, "ETag", eTag); // Required in 304.
			setDateHeader(header, "Expires", expires); // Postpone cache with 1 week.
			writeHeader(header, true);
			return;
		}


		// Validate request headers for resume ----------------------------------------------------

		// If-Match header should contain "*" or ETag. If not, then return 412.
		String ifMatch = m_headers.get("If-Match");
		if (ifMatch != null && !matches(ifMatch, eTag)) {
			setStatus(header, SC_PRECONDITION_FAILED);
			writeHeader(header, true);
			return;
		}

		// If-Unmodified-Since header should be greater than LastModified. If not, then return 412.
		long ifUnmodifiedSince = getDateHeader("If-Unmodified-Since");
		if (ifUnmodifiedSince != -1 && ifUnmodifiedSince + 1000 <= lastModified) {
			setStatus(header, SC_PRECONDITION_FAILED);
			writeHeader(header, true);
			return;
		}


		// Validate and process range -------------------------------------------------------------

		// Prepare some variables. The full Range represents the complete file.
		Range full = new Range(0, length - 1, length);
		List<Range> ranges = new ArrayList<Range>();

		// Validate and process Range and If-Range headers.
		String range = m_headers.get("Range");
		if (range != null) {

			// Range header should match format "bytes=n-n,n-n,n-n...". If not, then return 416.
			if (!range.matches("^bytes=\\d*-\\d*(,\\d*-\\d*)*$")) {
				setHeader(header, "Content-Range", "bytes */" + length); // Required in 416.
				setStatus(header, SC_REQUESTED_RANGE_NOT_SATISFIABLE);
				writeHeader(header, true);
				return;
			}

			// If-Range header should either match ETag or be greater then LastModified. If not,
			// then return full file.
			String ifRange = m_headers.get("If-Range");
			if (ifRange != null && !ifRange.equals(eTag)) {
				try {
					long ifRangeTime = getDateHeader("If-Range"); // Throws IAE if invalid.
					if (ifRangeTime != -1 && ifRangeTime + 1000 < lastModified) {
						ranges.add(full);
					}
				} catch (IllegalArgumentException ignore) {
					ranges.add(full);
				}
			}

			// If any valid If-Range header, then process each part of byte range.
			if (ranges.isEmpty()) {
				for (String part : range.substring(6).split(",")) {
					// Assuming a file with length of 100, the following examples returns bytes at:
					// 50-80 (50 to 80), 40- (40 to length=100), -20 (length-20=80 to length=100).
					long start = sublong(part, 0, part.indexOf("-"));
					long end = sublong(part, part.indexOf("-") + 1, part.length());

					if (start == -1) {
						start = length - end;
						end = length - 1;
					} else if (end == -1 || end > length - 1) {
						end = length - 1;
					}

					// Check if Range is syntactically valid. If not, then return 416.
					if (start > end) {
						setHeader(header, "Content-Range", "bytes */" + length); // Required in 416.
						setStatus(header, SC_REQUESTED_RANGE_NOT_SATISFIABLE);
						writeHeader(header, true);
						return;
					}

					// Add range.
					ranges.add(new Range(start, end, length));
				}
			}
		}


		// Prepare and initialize response --------------------------------------------------------

		// Initialize response.
		setHeader(header, "Content-Disposition", contentDisposition + "; filename=\"" + fileName + "\"");
		//setHeader(header, "Content-Type", "video/avi");
		setHeader(header, "Accept-Ranges", "bytes");
		setHeader(header, "ETag", eTag);
		setDateHeader(header, "Last-Modified", lastModified);
		setDateHeader(header, "Expires", expires);


		// Send requested file (part(s)) to client ------------------------------------------------

		// Prepare streams.
		RandomAccessFile input = null;

		try {
			// Open streams.
			input = new RandomAccessFile(file, "r");

			if (ranges.isEmpty() || ranges.get(0) == full) {

				// Return full file.
				setStatus(header, SC_OK); // 200
				Range r = full;
				setHeader(header, "Content-Range", "bytes " + r.start + "-" + r.end + "/" + r.total);
				setHeader(header, "Content-Length", String.valueOf(r.length));
				writeHeader(header, true);

				// Copy full range.
				copy(input, m_outStream, r.start, r.length);

			} else if (ranges.size() == 1) {

				// Return single part of file.
				Range r = ranges.get(0);
				setHeader(header, "Content-Range", "bytes " + r.start + "-" + r.end + "/" + r.total);
				setHeader(header, "Content-Length", String.valueOf(r.length));
				setStatus(header, SC_PARTIAL_CONTENT); // 206.
				writeHeader(header, true);

				// Copy single part range.
				copy(input, m_outStream, r.start, r.length);

			} else {

				// Return multiple parts of file.
				setHeader(header, "Content-Type", "multipart/byteranges; boundary=" + MULTIPART_BOUNDARY);
				setStatus(header, SC_PARTIAL_CONTENT); // 206.
				writeHeader(header, false);


				// Copy multi part range.
				for (Range r : ranges) {
					// Add multipart boundary and header fields for every range.
					String text = "\n--" + MULTIPART_BOUNDARY + "\n" + 
							"Content-Range: bytes " + r.start + "-" + r.end + "/" + r.total + "\n\n";
					m_outStream.write(text.getBytes("UTF-8"));

					// Copy single part range of multi part range.
					copy(input, m_outStream, r.start, r.length);
				}

				// End with multipart boundary.
				m_outStream.write( ("\n--" + MULTIPART_BOUNDARY + "--\n").getBytes("UTF-8") );
			}
		} finally {
			// Gently close streams.
			close(input);
		}
	}

	private static final int DEFAULT_BUFFER_SIZE = 10240; // ..bytes = 10KB.
	private static final long DEFAULT_EXPIRE_TIME = 604800000L; // ..ms = 1 week.
	private static final String MULTIPART_BOUNDARY = "MULTIPART_BYTERANGES";
	private static final String SC_OK = "HTTP/1.1 200 OK";
	private static final String SC_PARTIAL_CONTENT = "HTTP/1.1 206 Partial Content";
	private static final String SC_NOT_MODIFIED = "HTTP/1.1 304 Not Modified";
	private static final String SC_PRECONDITION_FAILED = "HTTP/1.1 412 Precondition Failed";
	private static final String SC_REQUESTED_RANGE_NOT_SATISFIABLE = "HTTP/1.1 416 Request Range Not Satisfiable";

	/**
	 * Returns true if the given match header matches the given value.
	 * @param matchHeader The match header.
	 * @param toMatch The value to be matched.
	 * @return True if the given match header matches the given value.
	 */
	private static boolean matches(String matchHeader, String toMatch) {
		String[] matchValues = matchHeader.split("\\s*,\\s*");
		Arrays.sort(matchValues);
		return Arrays.binarySearch(matchValues, toMatch) > -1
				|| Arrays.binarySearch(matchValues, "*") > -1;
	}

	/**
	 * Returns a substring of the given string value from the given begin index to the given end
	 * index as a long. If the substring is empty, then -1 will be returned
	 * @param value The string value to return a substring as long for.
	 * @param beginIndex The begin index of the substring to be returned as long.
	 * @param endIndex The end index of the substring to be returned as long.
	 * @return A substring of the given string value as long or -1 if substring is empty.
	 */
	private static long sublong(String value, int beginIndex, int endIndex) {
		String substring = value.substring(beginIndex, endIndex);
		return (substring.length() > 0) ? Long.parseLong(substring) : -1;
	}

	/**
	 * Copy the given byte range of the given input to the given output.
	 * @param input The input to copy the given range to the given output for.
	 * @param output The output to copy the given range from the given input for.
	 * @param start Start of the byte range.
	 * @param length Length of the byte range.
	 * @throws IOException If something fails at I/O level.
	 */
	private static void copy(RandomAccessFile input, OutputStream output, long start, long length)
			throws Exception
			{
		byte[] buffer = new byte[DEFAULT_BUFFER_SIZE];
		int read;
		String name = Thread.currentThread().getName();
		long prevTime = System.nanoTime();
		long currentTime;
		long prevSize = 0;
		long currentSize = 0;

		if (input.length() == length) {
			// Write full range.
			while ((read = input.read(buffer)) > 0) {
				output.write(buffer, 0, read);
				currentSize += read;
				currentTime = System.nanoTime();
				if (currentTime > (prevTime + 5 * 1000000000L))
				{
					setStatus(name, prevSize, currentSize, length, prevTime, currentTime);
					prevSize = currentSize;
					prevTime = currentTime;
				}
				//Thread.sleep(1000);	// Limit speed for testing.
			}
		} else {
			// Write partial range.
			input.seek(start);
			long toRead = length;

			while ((read = input.read(buffer)) > 0) {
				if ((toRead -= read) > 0) {
					output.write(buffer, 0, read);
					currentSize += read;
					currentTime = System.nanoTime();
					if (currentTime > (prevTime + 5 * 1000000000L))
					{
						setStatus(name, prevSize, currentSize, length, prevTime, currentTime);
						prevSize = currentSize;
						prevTime = currentTime;
					}
					//Thread.sleep(1000);	// Limit speed for testing.
				} else {
					output.write(buffer, 0, (int) toRead + read);
					break;
				}
			}
		}
			}

	/**
	 * Close the given resource.
	 * @param resource The resource to be closed.
	 */
	private static void close(Closeable resource) {
		if (resource != null) {
			try {
				resource.close();
			} catch (IOException ignore) {
				// Ignore IOException. If you want to handle this anyway, it might be useful to know
				// that this will generally only be thrown when the client aborted the request.
			}
		}
	}

	// Inner classes ------------------------------------------------------------------------------

	/**
	 * This class represents a byte range.
	 */
	protected class Range {
		long start;
		long end;
		long length;
		long total;

		/**
		 * Construct a byte range.
		 * @param start Start of the byte range.
		 * @param end End of the byte range.
		 * @param total Total length of the byte source.
		 */
		public Range(long start, long end, long total) {
			this.start = start;
			this.end = end;
			this.length = end - start + 1;
			this.total = total;
		}

	}
}