/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.util.*;
import java.util.Map.Entry;
import java.text.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;

import java.io.*;

class GuideStore
{
	private SortedMap<String, GuideChannel> m_programs = null;

	private Date m_maxEntry = new Date();
	private Date m_minEntry = new Date();
	private Vector<ChannelMapGroup> m_channelMapsInternal = new Vector<ChannelMapGroup>();
	private Vector<ChannelMapGroup> m_channelMapsCustom = new Vector<ChannelMapGroup>();
	private Vector<ChannelMapGroup> m_channelMaps = null;
	private Set<String> m_categorySet = new HashSet<String>();
	private Vector<AddConflictDetails> m_conflictList = new Vector<AddConflictDetails>();
	private HashMap<String, HashMap<String, ExclusionItem> > m_epgExclusionMap = new HashMap<String, HashMap<String, ExclusionItem> >();
	PriorityQueue<EpgLoadTime> m_internalEpgScheduleQueue = new PriorityQueue<EpgLoadTime>();

	private static GuideStore instance = null;

	private long m_epgLoadTime = Long.MAX_VALUE;
	private long m_deletedOldEpgTime = Long.MAX_VALUE;
	private boolean m_bReloadInternalEpg = false;

	private GuideStore()
	{
		loadEpg();
		loadChannelMap();
		loadCategories();
		loadEpgExclusion();
		loadInternalEpgSchedule();

		setInternalEpgChannelMapping();		
	}

	public static GuideStore getInstance()
	{
		synchronized (DataStore.class)
		{
			if (instance == null)
			{
				instance = new GuideStore();
				return instance;
			}
			else
			{
				return instance;
			}
		}
	}

	public String toString()
	{
		return "Programs Loaded:" + m_programs.size() + " Channel Maps Loads:" + m_channelMapsCustom.size() + " Categories Loaded:" + m_categorySet.size();
	}

	public Date getMaxEntry()
	{
		return m_maxEntry;
	}

	public Date getMinEntry()
	{
		return m_minEntry;
	}

	private Set<String> getCategories()
	{
		return m_categorySet;
	}

	private void setCategories(Set<String> cats)
	{
		m_categorySet = cats;
	}

	public int numberOfCategories()
	{
		return m_categorySet.size();
	}

	public Date getEPGmaxDate()
	{
		return getMaxEntry();
	}

	public Date getEPGminDate()
	{
		return getMinEntry();
	}   

	private HashMap<String, HashMap<String, ExclusionItem> > getEpgExclusionMap()
	{
		return m_epgExclusionMap;
	}

	private void setEpgExclusionMap(HashMap<String, HashMap<String, ExclusionItem> > exclusion)
	{
		m_epgExclusionMap = exclusion;
	}

	public boolean isInEpgExclusion(GuideItem guideItem, String channel)
	{
		HashMap<String, ExclusionItem> map = m_epgExclusionMap.get(channel);

		if (map != null)
		{
			ExclusionItem exc = map.get(guideItem.toString());

			if (exc != null)
			{
				return ( (exc.getName().equals(guideItem.getName().toUpperCase())) && 
						(exc.getStop().equals(guideItem.getStop())) );
			}
		}

		return false;
	}

	public void addEpgExclusion(GuideItem guideItem, String channel)
	{
		HashMap<String, ExclusionItem> map = m_epgExclusionMap.get(channel);
		if (map == null)
		{
			map = new HashMap<String, ExclusionItem>();
			m_epgExclusionMap.put(channel, map);
		}
		map.put(guideItem.toString(), 
				new ExclusionItem(guideItem.getName().toUpperCase(), guideItem.getStart(), guideItem.getStop()));

		guideItem.setAutoMark(false);
	}

	public boolean toggleEpgExclusion(GuideItem guideItem, String channel)
	{
		HashMap<String, ExclusionItem> map = m_epgExclusionMap.get(channel);

		if (map != null)
		{
			ExclusionItem exc = map.get(guideItem.toString());

			if (exc != null)
			{
				if ( (exc.getName().equals(guideItem.getName().toUpperCase())) && 
						(exc.getStop().equals(guideItem.getStop())) )
				{
					map.remove(guideItem.toString());
					if (map.isEmpty())
						m_epgExclusionMap.remove(channel);

					return false;
				}
			}
		}

		addEpgExclusion(guideItem, channel);
		return true;
	}

	public void renameEpgExclusionChannel(String channelOld, String channel)
	{
		HashMap<String, ExclusionItem> map = m_epgExclusionMap.get(channelOld);

		if (map != null)
		{
			m_epgExclusionMap.remove(channelOld);
			m_epgExclusionMap.put(channel, map);
		}

		saveEpgExclusion();
	}

	public Vector<AddConflictDetails> getAutoAddConflicts()
	{
		return getConflictList();
	}   

	public Vector<String[]> getEpgLinks(Calendar hiLight)
	{
		DataStore store = DataStore.getInstance();

		Date end = getMaxEntry();
		Date start = getMinEntry();

		if (start == null || end == null)
			return new Vector<String[]>();

			Vector<String[]> data = new Vector<String[]>();
			Calendar calStart = Calendar.getInstance();
			calStart.setTime(start);
			Calendar cal = (Calendar) calStart.clone();
			cal.set(Calendar.HOUR_OF_DAY, 6);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 0);
			if (cal.after(calStart))
				cal.add(Calendar.DAY_OF_MONTH, -1);

			Calendar now = Calendar.getInstance();

			while (cal.getTime().getTime() <= end.getTime())
			{
				String[] link = new String[3];
				link[0] = "year=" + cal.get(Calendar.YEAR) + "&month=" + (cal.get(Calendar.MONTH) + 1) + "&day=" + cal.get(Calendar.DATE);

				String dayName = (String) store.dayName.get(new Integer(cal.get(Calendar.DAY_OF_WEEK)));
				if(now.get(Calendar.YEAR) == cal.get(Calendar.YEAR) && 
						now.get(Calendar.MONTH) == cal.get(Calendar.MONTH) &&
						now.get(Calendar.DATE) == cal.get(Calendar.DATE))
				{
					dayName = "Today";
				}

				if(hiLight != null && 
						hiLight.get(Calendar.YEAR) == cal.get(Calendar.YEAR) && 
						hiLight.get(Calendar.MONTH) == cal.get(Calendar.MONTH) &&
						hiLight.get(Calendar.DATE) == cal.get(Calendar.DATE))
				{
					link[1] = "(" + dayName + ")";
				}
				else
				{
					link[1] = dayName;
				}

				link[2] = //store.dayName.get(new Integer(cal.get(Calendar.DAY_OF_WEEK))) 
						//+ ", " +
						cal.get(Calendar.DATE) + " " +
						store.monthNameShort.get(new Integer(cal.get(Calendar.MONTH)));
				// + ", " +
				//cal.get(Calendar.YEAR);

				data.add(link);
				cal.add(Calendar.HOUR, 24);
			}

			return data;
	}   

	@SuppressWarnings("unchecked")
	private void loadCategories()
	{
		try
		{
			FileInputStream fis = new FileInputStream(DataStore.m_saveFolderData + "Categories.sof");
			ObjectInputStream ois = new ObjectInputStream(fis);
			Set<String> cats = (HashSet<String>) ois.readObject();
			ois.close();

			setCategories(cats);

			DVScheduler.log("Categories.sof loaded (" + cats.size() + ")");
		}
		catch (Exception e)
		{
			setCategories(new HashSet<String>());
			DVScheduler.log("Error loading Categories.sof, starting with blank category list");
			e.printStackTrace();
		}
	}

	public void setAndSaveCategories()
	{
		if(m_categorySet == null)
		{
			m_categorySet = new HashSet<String>();
		}
		else
			m_categorySet.clear();

		Iterator<GuideChannel> iter = m_programs.values().iterator();
		while (iter.hasNext())
		{
			GuideChannel guideChannel = iter.next();
			guideChannel.getCategories(m_categorySet);
		}

		try
		{
			FileOutputStream fos = new FileOutputStream(DataStore.m_saveFolderData + "Categories.sof");
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(getCategories());
			oos.close();
			DVScheduler.log("Categories.sof saved.");
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}   

	private void deleteOldEpgExclusion()
	{
		calculateDateRange();
		Date min = getMinEntry();
		if (min == null)
			return;

		int countOld = 0;
		String[] channels = m_epgExclusionMap.keySet().toArray(new String[0]);
		for (int x = 0; x < channels.length; x++)
		{
			HashMap<String, ExclusionItem> exc = m_epgExclusionMap.get(channels[x]);

			String[] startTimes = exc.keySet().toArray(new String[0]);
			for (int y = 0; y < startTimes.length; y++)
			{
				Date stop = exc.get(startTimes[y]).getStop();

				if (stop.before(min))
				{
					exc.remove(startTimes[y]);
					countOld++;
					if (exc.isEmpty())
					{
						m_epgExclusionMap.remove(channels[x]);
						break;
					}
				}
			}
		}

		if (countOld > 0)
		{
			DVScheduler.log("Removed " + countOld + " old EPG exclusions");
			saveEpgExclusion();
		}
	}

	@SuppressWarnings("unchecked")
	private void loadEpgExclusion()
	{
		try
		{
			FileInputStream fis = new FileInputStream(DataStore.m_saveFolderData + "EpgExclusion.sof");
			ObjectInputStream ois = new ObjectInputStream(fis);
			HashMap<String, HashMap<String, ExclusionItem> > exclusion = (HashMap<String, HashMap<String, ExclusionItem> >) ois.readObject();
			ois.close();

			setEpgExclusionMap(exclusion);

			DVScheduler.log("EpgExclusion.sof loaded (" + exclusion.size() + " channel maps)");

			deleteOldEpgExclusion();
		}
		catch (Exception e)
		{
			setEpgExclusionMap(new HashMap<String, HashMap<String, ExclusionItem> >());
			DVScheduler.log("Error loading EpgExclusion.sof, starting with blank EPG exclusion list");
			e.printStackTrace();
		}
	}

	public void saveEpgExclusion()
	{
		try
		{
			FileOutputStream fos = new FileOutputStream(DataStore.m_saveFolderData + "EpgExclusion.sof");
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(getEpgExclusionMap());
			oos.close();
			DVScheduler.log("EpgExclusion.sof saved.");
		}
		catch (Exception e)
		{
			DVScheduler.log("Error saving EpgExclusion.sof");
			e.printStackTrace();
		}
	}

	@SuppressWarnings("unchecked")
	private void loadInternalEpgSchedule()
	{
		try
		{
			FileInputStream fis = new FileInputStream(DataStore.m_saveFolderData + "InternalEpgSchedule.sof");
			ObjectInputStream ois = new ObjectInputStream(fis);
			PriorityQueue<EpgLoadTime> queue = (PriorityQueue<EpgLoadTime>) ois.readObject();
			ois.close();

			m_internalEpgScheduleQueue = queue;
			DVScheduler.log("InternalEpgSchedule.sof loaded (" + queue.size() + " schedules)");
		}
		catch (Exception e)
		{
			DVScheduler.log("Error loading InternalEpgSchedule.sof, resetting schedule");
			setInternalEpgScheduleChannelChange();
			e.printStackTrace();
		}
	}

	public void saveInternalEpgSchedule()
	{
		try
		{
			FileOutputStream fos = new FileOutputStream(DataStore.m_saveFolderData + "InternalEpgSchedule.sof");
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(m_internalEpgScheduleQueue);
			oos.close();
			DVScheduler.log("InternalEpgSchedule.sof saved.");
		}
		catch (Exception e)
		{
			DVScheduler.log("Error saving InternalEpgSchedule.sof");
			e.printStackTrace();
		}
	}

	// nRunPostTask:
	//	Bit 1 according to system setting.
	//	Bit 2 run.
	private void runPostLoadActions(StringBuffer buff, int format, int caller, int nRunPostTaskAutoMark, int nRunPostTaskAutoAdd) throws Exception
	{
		if(format == 1)
			buff.append("Running post load action<br>\n");
		else
			buff.append("Running post load action\n");

		DataStore store = DataStore.getInstance();

		String titleCase = store.getProperty("guide.postScanChangeTitleCase", "unchanged").trim();
		if (titleCase.equals("lower"))
		{
			Iterator<GuideChannel> iter = m_programs.values().iterator();
			while (iter.hasNext())
			{
				iter.next().setToLowerCase();
			}
		}
		else if (titleCase.equals("upper"))
		{
			Iterator<GuideChannel> iter = m_programs.values().iterator();
			while (iter.hasNext())
			{
				iter.next().setToUpperCase();
			}
		}
		else if (titleCase.equals("title"))
		{
			Iterator<GuideChannel> iter = m_programs.values().iterator();
			while (iter.hasNext())
			{
				iter.next().setToTitleCase();
			}
		}

		boolean bRun = false;
		if ((nRunPostTaskAutoAdd & 0x02) > 0)
			bRun = true;
		else if ((nRunPostTaskAutoAdd & 0x01) > 0)
		{
			String postScan = store.getProperty("guide.postScanAutoAdd", "1");
			String postScanRun = store.getProperty("guide.postScanAutoAddRun", "All");

			bRun = postScan.equals("1") && (postScanRun.equals("All") || 
					(postScanRun.equals("Manual Data Load") && (caller == 0)) || 
					(postScanRun.equals("Scheduled Data Load") && (caller == 1)));
		}

		if ( bRun )
		{
			store.removeSchedulesOfTypeEpg(buff, format);
			addEpgMatches(buff, format, false);
			saveEpg();
			store.saveSchedules(null);
		}
		else
		{
			if ((nRunPostTaskAutoMark & 0x02) > 0)
				bRun = true;
			else if ((nRunPostTaskAutoMark & 0x01) > 0)
			{
				String postScan = store.getProperty("guide.postScanAutoMark", "1");
				String postScanRun = store.getProperty("guide.postScanAutoMarkRun", "All");

				bRun = postScan.equals("1") && (postScanRun.equals("All") || 
						(postScanRun.equals("Manual Data Load") && (caller == 0)) || 
						(postScanRun.equals("Scheduled Data Load") && (caller == 1)));
			}

			if ( bRun )
			{
				addEpgMatches(buff, format, true);
				saveEpg();
			}
		}

		if(format == 1)
			buff.append("Post load actions finished<br>\n");
		else
			buff.append("Post load actions finished\n");      

	}

	private void runTask(TaskCommand taskCommand, boolean waitForExit, StringBuffer buff, int format)
	{  
		DVScheduler.log("Running task (" + taskCommand.getName() + ")\n");

		if (format == 1)
			buff.append("Running task (" + taskCommand.getName() + ")<br>\n");
		else
			buff.append("Running task (" + taskCommand.getName() + ")\n");      

		if(taskCommand.getEnabled() == false)
		{
			DVScheduler.log("Task (" + taskCommand.getName() + ") is disabled.");
			if (format == 1)
				buff.append("Task (" + taskCommand.getName() + ") is disabled.<br>\n");
			else
				buff.append("Task (" + taskCommand.getName() + ") is disabled.\n");

			return;
		}

		String command = taskCommand.getCommand();
		boolean autoRem = taskCommand.getAutoRemove();
		int delay = taskCommand.getDelay();

		if(command != null && command.length() > 1)
		{
			DataStore store = DataStore.getInstance();
			command = store.replaceKeywords(command);

			if ( taskCommand.isConditional() && (waitForExit == false) )
			{
				DVScheduler.log("Queuing task (" + taskCommand.getName() + ")");
				TaskItem taskItem = new TaskItem(taskCommand, command, null);
				ThreadLock.getInstance().getLock();
				try
				{
					store.m_waitingTaskList.add(taskItem);
				}
				finally
				{
					ThreadLock.getInstance().releaseLock();
				}
			}
			else
			{
				try
				{
					runCommand(taskCommand, command, waitForExit, autoRem, delay, buff, format);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		}
	}

	private int runCommand(TaskCommand taskCommand, String command, boolean wait,
			boolean autoRem, int delay, StringBuffer buff, int format) throws Exception
			{
		DataStore store = DataStore.getInstance();
		long timeout = 60*60;

		if(delay > 0)
		{
			DVScheduler.log("The command has a delayed start of " + delay + " seconds");
			if (format == 1)
				buff.append("The command has a delayed start of " + delay + " seconds<br>\n");
			else
				buff.append("The command has a delayed start of " + delay + " seconds\n");
		}

		CommandWaitThread comThread = new CommandWaitThread(command, delay);
		Thread comWait = new Thread(Thread.currentThread().getThreadGroup(), comThread, comThread.getClass().getName());
		comWait.start();

		ThreadLock.getInstance().getLock();
		try
		{
			String key = new Long(new Date().getTime()).toString() + "-" + new Long(store.m_rand.nextLong()).toString();
			TaskItemThread taskItem = new TaskItemThread(key, comThread, command, autoRem);
			taskItem.m_allowSuspend = taskCommand.isAllowSuspend();
			Thread taskThread = new Thread(Thread.currentThread().getThreadGroup(), taskItem, taskItem.getClass().getName());
			taskThread.start();
			store.m_runningTaskList.put(key, taskItem);
		}
		finally
		{
			ThreadLock.getInstance().releaseLock();
		}
		if(wait == false)
		{
			return 1;
		}

		DVScheduler.log("Will wait for " + timeout + " seconds for it to complete");
		if (format == 1)
			buff.append("Will wait for " + timeout + " seconds for it to complete<br>\n");
		else
			buff.append("Will wait for " + timeout + " seconds for it to complete\n");

		long started = System.nanoTime();
		while (true)
		{
			if(comThread.isFinished())
			{
				DVScheduler.log("Command finished normally");
				if (format == 1)
					buff.append("Command finished normally<br>\n");
				else
					buff.append("Command finished normally\n");

				break;
			}
			else if((System.nanoTime() - started) > (1000000000L * timeout))
			{
				DVScheduler.log("Command timed out, killing it now");
				if (format == 1)
					buff.append("Command timed out, killing it now<br>\n");
				else
					buff.append("Command timed out, killing it now\n");

				comThread.stop();
				break;
			}
			DVScheduler.log("Waiting for command to finish.....");
			/*
			if (format == 1)
				buff.append("Waiting for command to finish.....<br>\n");
			else
				buff.append("Waiting for command to finish.....\n");
			 */

			Thread.sleep(5000);
		}

		DVScheduler.log("Task finished, Output of task follows");
		DVScheduler.log("*****************************");

		DVScheduler.log("Standard Out:\n" + comThread.getInputStream());
		DVScheduler.log("Standard Error:\n" + comThread.getErrorStream());

		DVScheduler.log("*****************************");

		if (format == 1)
		{
			buff.append("Task finished<br><br>\n\n");
		}
		else
		{
			buff.append("Task finished\n\n");
		}

		/*
		if (format == 1)
		{
			buff.append("Task finished, Output of task follows<br>\n");
			buff.append("<font size='-1' face='xyz, Courier New, Courier'>*****************************<br>\n");

			buff.append("Standard Out:<br>\n" + comThread.getInputStream().replace("\n", "<br>\n").replace(" ", "&nbsp;"));
			buff.append("<br>\nStandard Error:<br>\n" + comThread.getErrorStream().replace("\n", "<br>\n").replace(" ", "&nbsp;"));

			buff.append("<br>\n*****************************</font><br><br>\n\n");
		}
		else
		{
			buff.append("Task finished, Output of task follows\n");
			buff.append("*****************************\n");

			buff.append("Standard Out:\n" + comThread.getInputStream());
			buff.append("\nStandard Error:\n" + comThread.getErrorStream());

			buff.append("\n*****************************\n\n");
		}
		 */

		return 0;
			}

	public void loadXmltv(StringBuffer buff, int format, int caller) throws Exception
	{
		loadXmltv(buff, format, caller, 1, 1, 1);
	}

	// caller values:
	//	0 is manual data load.
	//	1 is scheduled data load.
	// nRunPreTask:
	//	0 don't run.
	//	1 according to system setting.
	//	2 run.
	public void loadXmltv(StringBuffer buff, int format, int caller, int nRunPreTask, int nRunPostTaskAutoMark, int nRunPostTaskAutoAdd) throws Exception
	{
		deleteOldEpg();
		DataStore store = DataStore.getInstance();

		boolean bRunPreTask = false;
		if (nRunPreTask == 2)
			bRunPreTask = true;
		else if (nRunPreTask == 1)
		{
			String preActionRun = store.getProperty("guide.preActionRun", "");
			if (preActionRun.equals("All"))
				bRunPreTask = true;
			else if ( (preActionRun.equals("Scheduled Data Load")) && (caller == 1) )
				bRunPreTask = true;
			else if ( (preActionRun.equals("Manual Data Load")) && (caller == 0) )
				bRunPreTask = true;
		}

		if (bRunPreTask)
		{
			String preTaskName = store.getProperty("Tasks.PreEPGloadTask", "");
			HashMap<String, TaskCommand> tasks = store.getTaskList();
			TaskCommand taskCommand = (TaskCommand) tasks.get(preTaskName);

			if (taskCommand != null)
			{
				/*
				if (format == 1)
					buff.append("Running pre-task (" + preTaskName + ")<br><br>\n");
				else
					buff.append("Running pre-task (" + preTaskName + ")\n\n");
				 */

				boolean bWhenNotCapturing = taskCommand.isWhenNotCapturing();
				int minIdleTime = taskCommand.getMinIdleTime();
				int maxConcurrentTask = taskCommand.getMaxConcurrentTask();
				CaptureDeviceList devList = CaptureDeviceList.getInstance();
				if ( (store.m_bSoonCapturing || !store.m_captureTasks.isEmpty() || (devList.getRunningDeviceCount() > 0)) && (bWhenNotCapturing || (minIdleTime > 0)) )
					bRunPreTask = false;

				if ( bRunPreTask && (maxConcurrentTask > 0) && (store.m_runningTaskList.size() >= maxConcurrentTask) )
					bRunPreTask = false;

				if ( bRunPreTask && (minIdleTime > 0) && ((store.m_nextScheduleTime - Calendar.getInstance().getTimeInMillis()) <= (minIdleTime*60*1000)) )
					bRunPreTask = false;

				if (bRunPreTask)
					runTask(taskCommand, true, buff, format);
				else
				{
					if (format == 1)
						buff.append("Cannot load EPG due to pre-task condition<br>\n");
					else
						buff.append("Cannot load EPG due to pre-task condition\n");

					return;
				}
			}
		}

		boolean bData = false;
		boolean bKeepOldData = (store.getPropertyInt("EPG.oldestItem", 72) > 0);
		try
		{
			String[] types = store.getProperty("guide.source.type", "Internal").split(",");
			int nType = 0;
			String location2 = "data/xmltv";
			for(int i = 0; i < types.length; i++)
			{
				if (types[i].trim().equals("File"))
				{
					nType = 0;
					location2 = store.getProperty("guide.source.file", "data/xmltv");
				}
				else if (types[i].trim().equals("Http"))
				{
					nType = 1;
					location2 = store.getProperty("guide.source.http", "");
				}
				else
					continue;

				String[] locations = DVScheduler.getArray(location2);
				String[] usernames = DVScheduler.getArray(store.getProperty("guide.source.http.usr", ""));
				String[] passwords = DVScheduler.getArray(store.getProperty("guide.source.http.pwd", ""));
				for(int x = 0; x < locations.length; x++)
				{
					String location = locations[x];
					if (location.isEmpty())
						continue;

					if (format == 1)
						buff.append("About to load XMLTV Data from " + location + "<br>\n");
					else
						buff.append("About to load XMLTV Data from " + location + "\n");

					GuideDataLoader loader = new GuideDataLoader();
					Vector<byte[]> dataFiles = null;
					if (nType == 0)
						dataFiles = loader.getDataFromFiles(location, buff, format);
					else if (nType == 1)
						dataFiles = loader.getDataFromURL(location, (x < usernames.length ? usernames[x] : ""), (x < passwords.length ? passwords[x] : ""));

					if(dataFiles.size() == 0)
					{
						if (format == 1)
							buff.append("No Data Found in " + location + "<br>\n");
						else
							buff.append("No Data Found in " + location + "\n");
					}
					else
					{
						//if (bData == false)
						//	clearGuideData();

						bData = true;
						for (int y = 0; y < dataFiles.size(); y++)
						{
							byte[] fileBytes = (byte[])dataFiles.get(y);

							if ( (fileBytes == null) || (fileBytes.length <= 0) )
							{
								if (format == 1)
									buff.append("Empty Content<br>\n");
								else
									buff.append("Empty Content\n");								
							}
							else
								loadXmlData(fileBytes, "", bKeepOldData);
						}
					}
				}
			}
		}
		catch (Exception e)
		{
			DVScheduler.log("Error Loading EPG DATA!");

			if (format == 1)
				buff.append("<br><strong>");

			buff.append("There was an error loading the XMLTV Data!");

			if (format == 1)
				buff.append("</strong><br>");

			buff.append("\n");

			ByteArrayOutputStream ba = new ByteArrayOutputStream();
			PrintWriter err = new PrintWriter(ba);
			e.printStackTrace(err);
			err.flush();

			if (format == 1)
				buff.append("<pre>");         

			buff.append(ba.toString());

			if (format == 1)
				buff.append("</pre>");    

			if (format == 1)
				buff.append("<br><br>");

			buff.append("\n");

			e.printStackTrace();
			return;
		}

		if (bData == false)
		{
			if (format == 1)
				buff.append("No Data Found!<br>\n");
			else
				buff.append("No Data Found!\n");

			return;
		}
		if (format == 1)
			buff.append("<br>");
		else
			buff.append("\n");

		// do post load actions
		StringBuffer actionBuff = new StringBuffer();
		runPostLoadActions(actionBuff, format, caller, nRunPostTaskAutoMark, nRunPostTaskAutoAdd);

		calculateDateRange();

		if (getMinEntry() != null && getMaxEntry() != null)
		{
			if (format == 1)
			{
				buff.append("Earliest Entry : " + getMinEntry().toString() + "<br>\n");
				buff.append("Latest Entry : " + getMaxEntry().toString() + "<br><br>\n");
			}
			else
			{
				buff.append("Earliest Entry : " + getMinEntry().toString() + "\n");
				buff.append("Latest Entry : " + getMaxEntry().toString() + "\n\n");
			}
		}

		Set<?> cats = getCategories();

		if (cats.size() > 0)
		{
			if (format == 1)
				buff.append("Number of Categories Found : " + String.valueOf(cats.size()) + "<br><br>\n");
			else
				buff.append("Number of Categories Found : " + String.valueOf(cats.size()) + "\n\n");
		}

		if (format == 1)
		{
			buff.append("Current EPG Data Set<br>\n");
			Iterator<Entry<String, GuideChannel>> iter = m_programs.entrySet().iterator();
			while (iter.hasNext())
			{
				Entry<String, GuideChannel> entry = iter.next();
				buff.append(entry.getKey() + " : " + entry.getValue().getSize() + "<br>\n");
			}
			buff.append("<br>Saving loaded EPG data<br><br>\n");
		}
		else
		{
			buff.append("Current EPG Data Set\n");
			Iterator<Entry<String, GuideChannel>> iter = m_programs.entrySet().iterator();
			while (iter.hasNext())
			{
				Entry<String, GuideChannel> entry = iter.next();
				buff.append(entry.getKey() + " : " + entry.getValue().getSize() + "<br>\n");
			}
			buff.append("\nSaving loaded EPG data\n\n");
		}

		m_epgLoadTime = new Date().getTime();

		buff.append(actionBuff);

		saveEpg();
		setAndSaveCategories();
		calculateDateRange();

		System.gc();

		boolean bRunPostTask = false;
		String postActionRun = store.getProperty("guide.postActionRun", "");
		if (postActionRun.equals("All"))
			bRunPostTask = true;
		else if ( (postActionRun.equals("Scheduled Data Load")) && (caller == 1) )
			bRunPostTask = true;
		else if ( (postActionRun.equals("Manual Data Load")) && (caller == 0) )
			bRunPostTask = true;

		if (bRunPostTask)
		{
			String postTaskName = store.getProperty("Tasks.PostEPGloadTask", "");
			HashMap<String, TaskCommand> tasks = store.getTaskList();
			TaskCommand taskCommand = (TaskCommand) tasks.get(postTaskName);

			if(taskCommand != null)
			{
				//runTask(taskCommand, true, buff, format);
				runTask(taskCommand, false, buff, format);
			}
		}

		adjustSchedulesToEpg();
	}   

	// This is for schedules manually added from EPG, or from Auto-Add that's already running.
	private void adjustSchedulesToEpg()
	{
		DataStore store = DataStore.getInstance();
		Iterator<ScheduleItem> iter = store.getScheduleList().values().iterator();
		ScheduleItem item = null;
		String epgChan = null;
		final int startBuffer = store.getPropertyInt("Schedule.buffer.start", 4);
		final int stopBuffer = store.getPropertyInt("Schedule.buffer.end", 5);
		final int epgBuffer = store.getPropertyInt("Schedule.buffer.EPG", 5);
		final long now = new Date().getTime();

		ThreadLock.getInstance().getLock();
		try {
			while (iter.hasNext())
			{
				item = iter.next();

				if ( (item.getCreatedFrom() == null) || (now >= item.getStop().getTime()) )
					continue;
				if ( !item.isTrackEpg() && ((item.getType() != ScheduleItem.EPG) || (item.getState() != ScheduleItem.RUNNING)) )
					continue;

				GuideItem itemProg = item.getCreatedFrom();
				epgChan = getEpgChannelIdFromMap(item.getChannel());
				if (epgChan == null)
					continue;
				GuideItem[] programs = getProgramsInc(item.getStart(), item.getStop(), epgChan);
				int maxOverLap = 0;
				GuideItem bestProg = null;

				for (int i=0; i<programs.length; i++)
				{
					GuideItem prog = programs[i];

					if ( itemProg.getName().equalsIgnoreCase(prog.getName()) && 
							itemProg.getSubName().equalsIgnoreCase(prog.getSubName()) )
					{
						int overLap = ScheduleItem.getOverlap(itemProg.getStart().getTime(), itemProg.getStop().getTime(), prog.getStart().getTime(), prog.getStop().getTime());
						if (overLap > maxOverLap)
						{
							maxOverLap = overLap;
							bestProg = prog;
						}
					}
				}

				if (bestProg != null)
				{
					item.log("Update start/stop time to new EPG");
					//DVScheduler.log("Updating EPG for schedule " + item.toString() + ": " + item.getName());
					try
					{
						if (item.getState() == ScheduleItem.WAITING)
						{
							item.setStart(bestProg.getStart().getTime() - startBuffer * 60000);
						}
						if ( (item.getState() == ScheduleItem.WAITING) || (item.getState() == ScheduleItem.RUNNING) )
						{
							item.setDuration( bestProg.getDuration()
									+ bestProg.getDuration() * epgBuffer / 60 
									+ startBuffer
									+ stopBuffer );
							item.setRunningStatus(new Date());
						}
						item.setCreatedFrom(bestProg);
					}
					catch (Exception e)
					{
					}
				}
				else
				{
					item.log("Schedule not found in updated EPG");
					//DVScheduler.log("EPG event not found for schedule " + item.toString() + ": " + item.getName());

					if ( !store.getProperty("email.captureWarning", "No").equals("Yes") )
						return;

					try
					{
						EmailSender sender = new EmailSender();
						sender.setSubject("DV Scheduler capture not found in updated EPG");
						sender.setBody(item.getLog());

						Thread mailThread = new Thread(Thread.currentThread().getThreadGroup(), sender, sender.getClass().getName());
						mailThread.start();
					}
					catch (Exception e)
					{
						e.printStackTrace();
					}
				}
			}
			store.saveSchedules(item);
		}
		finally
		{
			ThreadLock.getInstance().releaseLock();
		}
	}

	@SuppressWarnings("unchecked")
	public void loadEpg()
	{
		String loadFrom = DataStore.m_saveFolderData + "Epg.sof";
		try
		{
			FileInputStream fis = new FileInputStream(loadFrom);
			ObjectInputStream ois = new ObjectInputStream(fis);
			m_programs = (SortedMap<String, GuideChannel>) ois.readObject();
			ois.close();

			File file = new File(loadFrom);
			m_epgLoadTime = (file.exists() && file.isFile()) ? file.lastModified() : Long.MAX_VALUE;
			DVScheduler.log("Epg.sof loaded.");

			calculateDateRange();
		}
		catch (Exception e)
		{
			m_epgLoadTime = Long.MAX_VALUE;
			m_programs = new TreeMap<String, GuideChannel>();
			DVScheduler.log("Error loading Epg.sof, starting with empty EPG data set.");
			e.printStackTrace();
		}
	}   

	public void saveEpg()
	{
		String saveTo = DataStore.m_saveFolderData + "Epg.sof";
		try
		{
			FileOutputStream fos = new FileOutputStream(saveTo);
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(m_programs);
			oos.close();
			DVScheduler.log("Epg.sof saved.");
		}
		catch (Exception e)
		{
			DVScheduler.log("Error saving Epg.sof");
			e.printStackTrace();
		}
	}   

	@SuppressWarnings("unchecked")
	public void loadChannelMap()
	{
		String loadFrom = DataStore.m_saveFolderData + "ChannelMap.sof";
		try
		{
			FileInputStream fis = new FileInputStream(loadFrom);
			ObjectInputStream ois = new ObjectInputStream(fis);
			Vector<ChannelMapGroup> channelMap = (Vector<ChannelMapGroup>) ois.readObject();
			ois.close();
			if (channelMap.isEmpty())
				channelMap = new Vector<ChannelMapGroup>();
			else
			{
				Object object = channelMap.get(0);
				if (object instanceof ChannelMapGroup)
				{
					DVScheduler.log("ChannelMap.sof loaded.");

					setChannelMapCustom(channelMap);
				}
				else
				{
					FileInputStream fis2 = new FileInputStream(loadFrom);
					ObjectInputStream ois2 = new ObjectInputStream(fis2);
					Vector<String[]> channelMap2 = (Vector<String[]>) ois2.readObject();
					ois2.close();
					DVScheduler.log("Old ChannelMap.sof loaded.");

					for (int i=0; i<channelMap.size(); i++)
					{
						String[] map = channelMap2.get(i);
						// Old mapping structure array index:
						// 0: DV Scheduler channel name.
						// 1: EPG channel ID.
						// 2: EPG channel name.
						// 3: DV Scheduler channel setting.

						// DV Scheduler version 5.74 uses a new EPG channel ID, so convert.
						if ((map.length < 4) || map[3].isEmpty())
							continue;

						map[1] = map[3] + "+" + map[1];
						addChannelToCustomMap(map[0], map[1], map[2], map[3], -1);
					}
				}
			}
		}
		catch (Exception e)
		{
			DVScheduler.log("Error loading ChannelMap.sof, starting with no channel mapping.");
			e.printStackTrace();
		}
	}   

	public void saveChannelMap()
	{
		String saveTo = DataStore.m_saveFolderData + "ChannelMap.sof";
		try
		{
			FileOutputStream fos = new FileOutputStream(saveTo);
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			oos.writeObject(m_channelMapsCustom);
			oos.close();
			DVScheduler.log("ChannelMap.sof saved.");         
		}
		catch (Exception e)
		{
			e.printStackTrace();
			DVScheduler.log("Problem saving ChannelMap.sof");
		}
	}   

	protected ChannelMapGroup getChannelMapGroup(String channel)
	{
		for(int i=0; i<m_channelMaps.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(i);
			if (mapGroup.m_channelName.equals(channel))
				return mapGroup;
		}
		return null;
	}

	protected ChannelMapGroup getChannelMapGroupCustom(String channel)
	{
		for(int i=0; i<m_channelMapsCustom.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMapsCustom.get(i);
			if (mapGroup.m_channelName.equals(channel))
				return mapGroup;
		}
		return null;
	}

	public String getWsChannelFromMap(String epgChannelId)
	{
		if(epgChannelId == null)
			return null;

		for(int x = 0; x < m_channelMaps.size(); x++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(x);
			if (mapGroup.isExistEpgChannelId(epgChannelId))
				return mapGroup.m_channelName;
		}
		return null;
	}   

	private ChannelMap getBestChannelMap(ChannelMapGroup mapGroup)
	{
		List<ChannelMap> list = new LinkedList<ChannelMap>();

		long now = new Date().getTime();
		Vector<ChannelMap> maps = mapGroup.m_channelMaps;
		for (int j=0; j<maps.size(); j++)
		{
			ChannelMap map = maps.get(j);
			GuideChannel guideChannel = m_programs.get(map.m_epgChannelId);
			if ((guideChannel == null) || guideChannel.isEmpty() || guideChannel.isEmptyFutureEvents(now))
				continue;
			map.m_loadTime = guideChannel.getLoadTime();
			list.add(map);
		}

		if (list.isEmpty())
			return null;

		Comparator<ChannelMap> comparator = new Comparator<ChannelMap>() {
			public int compare(ChannelMap g1, ChannelMap g2) {
				if (g1.m_priority == g2.m_priority)
				{
					return (int) (g2.m_loadTime - g1.m_loadTime);
				}
				else
					return g1.m_priority - g2.m_priority;
			}
		};

		Collections.sort(list, comparator);
		//DVScheduler.log("Best channel: " + list.get(0).m_epgChannelId);
		return list.get(0);
	}

	public String getEpgChannelIdFromMap(String wsChannel)
	{
		if(wsChannel == null)
			return null;

		ChannelMapGroup mapGroup = getChannelMapGroup(wsChannel);
		if (mapGroup == null)
			return null;

		ChannelMap map = getBestChannelMap(mapGroup);
		return (map == null) ? null : map.m_epgChannelId;
	}   

	public int getChannelIndex(String wsChannel)
	{
		if(wsChannel == null)
			return -1;

		for(int x = 0; x < m_channelMaps.size(); x++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(x);
			if(mapGroup.m_channelName.equals(wsChannel))
				return x;
		}
		return -1;
	}   

	public String[] getCategoryStrings()
	{
		String[] out = {};
		int i = 0;
		if (m_categorySet.size() > 0)
		{
			out = new String[m_categorySet.size()];
			Iterator<String> it = m_categorySet.iterator();
			while (it.hasNext())
			{
				out[i] = it.next();
				i++;
			}
			Arrays.sort(out);
		}
		return out;
	}   

	public synchronized void calculateDateRange()
	{
		long minAll = Long.MAX_VALUE;
		long maxAll = Long.MIN_VALUE;

		if (!m_programs.isEmpty())
		{
			Iterator<GuideChannel> iter = m_programs.values().iterator();
			while (iter.hasNext())
			{
				GuideChannel item = iter.next();
				if (item.isEmpty())
					continue;
				long min = item.getMinStartTime();
				long max = item.getMaxStartTime();
				if (min < minAll)
					minAll = min;
				if (max > maxAll)
					maxAll = max;
			}
		}
		if (minAll == Long.MAX_VALUE)
			m_minEntry = null;
		else
			m_minEntry = new Date(minAll);
		if (maxAll == Long.MIN_VALUE)
			m_maxEntry = null;
		else
			m_maxEntry = new Date(maxAll);
	}

	public int loadXmlData(byte[] xmlFile, String name, boolean bKeepOldData) throws Exception
	{
		int found = 0;

		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
		ByteArrayInputStream reader = new ByteArrayInputStream(xmlFile);

		Document doc = docBuilder.parse(reader);
		NodeList root = null;

		// use the XMLTV format parser
		// get all TV nodes
		root = doc.getElementsByTagName("tv");
		if(root != null)
		{
			// get the node list from the first TV node found

			Node firstNode = root.item(0);
			if(firstNode == null)
			{
				DVScheduler.log("ERROR precessing XML data, first node of <tv> not found.");
				return found;
			}

			root = firstNode.getChildNodes();
			if(root == null)
			{
				DVScheduler.log("ERROR precessing XML data, no children of first node.");
				return found;
			}

			SortedMap<String, GuideChannel> programs = new TreeMap<String, GuideChannel>();
			HashMap< String, ChannelEpgScan > channels = loadChannels(root);
			found = recurseXMLData(root, channels, name, programs);
			if (found <= 0)
				return found;

			if (bKeepOldData)
			{
				Iterator<Entry<String, GuideChannel>> iter = programs.entrySet().iterator();
				while (iter.hasNext())
				{
					Entry<String, GuideChannel> entry = iter.next();
					GuideChannel guideChannel = entry.getValue();
					String key = entry.getKey();
					GuideChannel guideChannel2 = m_programs.get(key);
					if (guideChannel2 != null)
					{
						guideChannel.addOldItems(guideChannel2);
					}
				}
			}
			m_programs.putAll(programs);
		}

		return found;
	}

	private HashMap< String, ChannelEpgScan > loadChannels(NodeList root)
	{
		HashMap< String, ChannelEpgScan > channels = new HashMap< String, ChannelEpgScan >();

		Node thisNode = null;

		for (int x = 0; x < root.getLength(); x++)
		{
			thisNode = root.item(x);
			if(thisNode == null)
				break;

			if(thisNode.getNodeName().equals("channel"))
				parseChannel(thisNode, channels);
		}

		return channels;
	}

	private String getNodeValue(Node node)
	{
		if(node != null && node.getFirstChild() != null)
			return node.getFirstChild().getNodeValue();
		else
			return "";
	}

	private String getNodeString(Node prog, String nodename, String defaultValue)
	{
		String out = defaultValue;
		try
		{
			Node tmp = getChildByType(prog, nodename);
			if(tmp != null && tmp.getFirstChild() != null)
			{
				out = tmp.getFirstChild().getNodeValue();
			}
			tmp = null;
		}
		catch (Exception e)
		{
		}
		return out;
	}

	private String getNodeStrings(Node prog, String nodename, String defaultValue)
	{
		Node tmp = getChildByType(prog, nodename);
		if (tmp == null)
			return defaultValue;

		NodeList itemNodes = tmp.getChildNodes();
		if (itemNodes.getLength() <= 0)
			return defaultValue;

		Node itemNode = null;
		String out = itemNodes.item(0).getNodeValue();
		for(int y = 1; y < itemNodes.getLength(); y++)
		{
			itemNode = itemNodes.item(y);
			out += "\n" + itemNode.getNodeValue();
		}
		return out;
	}

	public boolean getNodeExists(Node prog, String name)
	{
		Node tmp = null;
		try
		{
			tmp = getChildByType(prog, name);
			if(tmp != null)
				return true;
		}
		catch (Exception e)
		{
		}
		return false;
	}

	public void clearGuideData()
	{
		m_programs = new TreeMap<String, GuideChannel>();
		m_categorySet = new HashSet<String>();
	}

	public GuideItem[] getProgramsForChannel(String epgChannelId)
	{
		if (epgChannelId == null)
			return new GuideItem[0];

		GuideChannel fullList = m_programs.get(epgChannelId);
		if(fullList == null)
		{
			return new GuideItem[0];
		}

		return fullList.getItems();
	}

	public int getProgramSizeForEpgChannel(String epgChannelId)
	{
		if (epgChannelId == null)
			return 0;

		GuideChannel fullList = m_programs.get(epgChannelId);
		if(fullList == null)
		{
			return 0;
		}

		return fullList.getSize();
	}

	public GuideItem[] getPrograms(Date start, Date end, String channel)
	{
		if (channel == null)
			return new GuideItem[0];

		GuideChannel fullList = m_programs.get(channel);
		if(fullList == null)
		{
			return new GuideItem[0];
		}

		return fullList.getItems(start, end);
	}

	public GuideItem[] getProgramsInc(Date start, Date end, String channel)
	{
		if (channel == null)
			return new GuideItem[0];

		GuideChannel fullList = m_programs.get(channel);
		if(fullList == null)
		{
			return new GuideItem[0];
		}

		return fullList.getItemsInc(start, end);
	}

	public GuideItem getProgram(String channel, String id)
	{
		if ((channel == null) || (id == null))
			return null;

		GuideChannel fullList = m_programs.get(channel);
		if(fullList == null)
			return null;

		return (GuideItem) fullList.getItem(id);
	}

	public GuideChannel getGuideChannel(String epgChannelId)
	{
		return m_programs.get(epgChannelId);
	}

	public int recurseXMLData(NodeList nl, HashMap< String, ChannelEpgScan > channels, String name, SortedMap<String, GuideChannel> programs)
	{
		DataStore store = DataStore.getInstance();
		String[] languages2 = store.getProperty("guide.language.priority", "").split(",");
		String maxLanguage2 = store.getProperty("guide.language.max", "");
		int maxLanguage = maxLanguage2.isEmpty() ? Integer.MAX_VALUE : Integer.parseInt(maxLanguage2);
		Map<String, Integer> languages = new HashMap<String, Integer>();
		for (int i=0; i<languages2.length; i++)
		{
			languages2[i] = languages2[i].trim();
			if (languages2[i].length() < 2)
				continue;
			languages.put(languages2[i].trim().substring(0, 2).toLowerCase(), i);
		}
		int found = 0;

		Node thisNode = null;
		for (int x = 0; x < nl.getLength(); x++)
		{
			thisNode = nl.item(x);
			if(thisNode == null)
				break;

			if(thisNode.getNodeName().equals("programme"))
			{
				if(parseProgram(thisNode, channels, name, programs, languages, maxLanguage))
					found++;
			}
		}

		return found;
	}

	public void moveChannel(int id, boolean dir)
	{
		int dest = 0;

		if(dir)
			dest = id + 1;
		else
			dest = id - 1;

		if(dest >= 0 && dest < m_channelMapsCustom.size())
		{
			ChannelMapGroup obj = m_channelMapsCustom.remove(id);

			if(dir)
				m_channelMapsCustom.add(id + 1, obj);
			else
				m_channelMapsCustom.add(id - 1, obj);

			try
			{
				saveChannelMap();
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
	}

	public void deleteChannels(Vector<Integer> indexes)
	{
		if ( (indexes == null) || indexes.isEmpty() )
			return;

		for (int i=0; i<indexes.size(); i++)
		{
			int index = indexes.get(i) - i;
			if ( (index >= 0) && (index < m_channelMapsCustom.size()) )
			{
				m_channelMapsCustom.remove(index);
			}
		}

		try
		{
			saveChannelMap();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	public void deleteAllChannels() throws Exception
	{
		m_channelMapsCustom.clear();
		saveChannelMap();
	}

	public void moveChannelsBeforeChannel(int beforeIndex, Vector<Integer> indexes)
	{
		if (beforeIndex < 0)
			beforeIndex = m_channelMapsCustom.size();

		Vector<ChannelMapGroup> maps = new Vector<ChannelMapGroup>();
		for (int i=0; i<indexes.size(); i++)
		{
			int index = indexes.get(i);
			if ( (index >= 0) && (index < m_channelMapsCustom.size()) )
				maps.add(m_channelMapsCustom.get(index));
		}
		for (int i=0; i<indexes.size(); i++)
		{
			int index = indexes.get(i) - i;
			if ( (index >= 0) && (index < m_channelMapsCustom.size()) )
			{
				m_channelMapsCustom.remove(index);
				if (index < beforeIndex)
					beforeIndex--;
			}
		}

		if ( (beforeIndex >= 0) && (beforeIndex <= m_channelMapsCustom.size()) )
		{
			m_channelMapsCustom.addAll(beforeIndex, maps);

			try
			{
				saveChannelMap();
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
	}

	public String[] getChannelList()
	{
		return (String[]) m_programs.keySet().toArray(new String[0]);
	}

	public Vector<String[]> getOldChannelMap()
	{
		Vector<String[]> oldChannelMap = new Vector<String[]>();
		for(int i=0; i<m_channelMaps.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(i);
			ChannelMap map = getBestChannelMap(mapGroup);
			if (map == null)
				continue;
			String[] oldMapping = { mapGroup.m_channelName, map.m_epgChannelId };
			oldChannelMap.add(oldMapping);
		}
		return oldChannelMap;
	}

	public Vector<ChannelMapGroup> getChannelMap()
	{
		return m_channelMaps;
	}

	public Vector<ChannelMapGroup> getChannelMapCustom()
	{
		return m_channelMapsCustom;
	}

	public void setChannelMapCustom(Vector<ChannelMapGroup> chanMap)
	{
		m_channelMapsCustom = chanMap;
	}

	public void deleteChannelMapCustom(int index, int subIndex)
	{
		if (index < 0 || index >= m_channelMapsCustom.size())
			return;
		ChannelMapGroup mapGroup = m_channelMapsCustom.get(index);

		if (subIndex < 0 || subIndex >= mapGroup.m_channelMaps.size())
			return;
		mapGroup.m_channelMaps.remove(subIndex);
		if (mapGroup.m_channelMaps.isEmpty())
			m_channelMapsCustom.remove(index);
		saveChannelMap();
	}

	public void deleteChannelMapCustom(int index)
	{
		if (index < 0 || index >= m_channelMapsCustom.size())
			return;

		m_channelMapsCustom.remove(index);
		saveChannelMap();
	}

	public void deleteChannelMapCustom(String channelName)
	{
		for(int i=0; i<m_channelMapsCustom.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMapsCustom.get(i);
			if (mapGroup.m_channelName.equals(channelName))
			{
				m_channelMapsCustom.remove(i);
				saveChannelMap();
				return;
			}
		}		
	}

	public void updateChannelMapCustom(String oldChannelName, String newChannelName)
	{
		ChannelMapGroup mapGroup = getChannelMapGroupCustom(oldChannelName);
		if (mapGroup == null)
			return;
		mapGroup.m_channelName = newChannelName;
		saveChannelMap();
	}

	public void updateChannelMapCustom(String channelName, String epgChannelId, int priority, int index, int subIndex)
	{
		if (index < 0 || index >= m_channelMapsCustom.size())
			return;
		ChannelMapGroup mapGroup = m_channelMapsCustom.get(index);

		if (subIndex < 0 || subIndex >= mapGroup.m_channelMaps.size())
			return;
		ChannelMap map = mapGroup.m_channelMaps.get(subIndex);
		map.m_epgChannelId = epgChannelId;
		map.m_priority = priority;
		mapGroup.sortChannelMaps();
		GuideChannel guideChannel = m_programs.get(epgChannelId);
		if (guideChannel == null)
			return;
		map.m_epgChannelName = guideChannel.getEpgChannelName();
		map.m_setting = guideChannel.getChannelSetting();

		if (channelName.equals(mapGroup.m_channelName))
			return;

		ChannelMapGroup mapGroup2 = getChannelMapGroupCustom(channelName);
		if (mapGroup2 == null)
		{
			mapGroup.m_channelName = channelName;
			return;
		}

		m_channelMapsCustom.remove(index);
		for (int i=0; i<mapGroup.m_channelMaps.size(); i++)
		{
			map = mapGroup.m_channelMaps.get(i);
			addChannelToCustomMap(channelName, map.m_epgChannelId, map.m_priority);
		}
	}

	public void addChannelToCustomMap(String channelName, String epgChannelId, int priority)
	{
		ChannelMapGroup mapGroup = getChannelMapGroupCustom(channelName);
		if (mapGroup == null)
		{
			mapGroup = new ChannelMapGroup();
			m_channelMapsCustom.add(mapGroup);
		}
		mapGroup.m_channelName = channelName;

		ChannelMap map = mapGroup.getChannelMap(epgChannelId);
		if (map == null)
		{
			map = new ChannelMap();
			mapGroup.m_channelMaps.add(map);
		}
		map.m_epgChannelId = epgChannelId;
		if (priority >= 0)
			map.m_priority = priority;
		mapGroup.sortChannelMaps();
		GuideChannel guideChannel = m_programs.get(epgChannelId);
		if (guideChannel == null)
			return;
		map.m_epgChannelName = guideChannel.getEpgChannelName();
		map.m_setting = guideChannel.getChannelSetting();
	}

	public void addChannelToCustomMap(String channelName, String epgChannelId, String epgChannelName, String channelSetting, int priority)
	{
		ChannelMapGroup mapGroup = getChannelMapGroupCustom(channelName);
		if (mapGroup == null)
		{
			mapGroup = new ChannelMapGroup();
			m_channelMapsCustom.add(mapGroup);
		}
		mapGroup.m_channelName = channelName;

		ChannelMap map = mapGroup.getChannelMap(epgChannelId);
		if (map == null)
		{
			map = new ChannelMap();
			mapGroup.m_channelMaps.add(map);
		}
		map.m_epgChannelId = epgChannelId;
		if (priority >= 0)
			map.m_priority = priority;
		mapGroup.sortChannelMaps();
		map.m_epgChannelName = epgChannelName;
		map.m_setting = channelSetting;
	}

	public Date parseDate(String newDate) throws NumberFormatException
	{
		String dateFormat = "";
		if(newDate.indexOf(" +") > -1 || newDate.indexOf(" -") > -1)
		{
			dateFormat = "yyyyMMddHHmmss Z";
		}
		else
		{
			dateFormat = "yyyyMMddHHmmss";
		}

		SimpleDateFormat df = new SimpleDateFormat(dateFormat);

		Date parsedDate = null;
		try
		{
			parsedDate = df.parse(newDate);
		}
		catch (Exception e)
		{
			e.printStackTrace();
			parsedDate = new Date(0);
		}

		return parsedDate;
	}

	private class ChannelEpgScan
	{
		int m_originalNetworkId = -2;
		int m_transportStreamId = -2;
		int m_programId = -2;
		int m_sourceId = -2;
		int m_originalNetworkId2 = -2;
		int m_transportStreamId2 = -2;
		int m_programId2 = -2;
		String m_displayName = null;
	}

	private void parseChannel(Node chan, HashMap< String, ChannelEpgScan > channels)
	{
		ChannelEpgScan epgChannel = new ChannelEpgScan();
		NamedNodeMap attMap = chan.getAttributes();
		Node node = attMap.getNamedItem("id");
		String id = node.getNodeValue();

		if ((id == null) || id.isEmpty())
			return;

		try
		{
			node = attMap.getNamedItem("originalNetworkId");
			if (node != null)
				epgChannel.m_originalNetworkId = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("transportStreamId");
			if (node != null)
				epgChannel.m_transportStreamId = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("programId");
			if (node != null)
				epgChannel.m_programId = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("originalNetworkId2");
			if (node != null)
				epgChannel.m_originalNetworkId2 = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("transportStreamId2");
			if (node != null)
				epgChannel.m_transportStreamId2 = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("programId2");
			if (node != null)
				epgChannel.m_programId2 = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			node = attMap.getNamedItem("sourceId");
			if (node != null)
				epgChannel.m_sourceId = Integer.parseInt(node.getNodeValue());
		}
		catch(Exception e)
		{
		}

		try
		{
			Node nameNode = getChildByType(chan, "display-name");
			if (nameNode != null && nameNode.getFirstChild() != null)
				epgChannel.m_displayName = nameNode.getFirstChild().getNodeValue();
		}
		catch(Exception e)
		{
		}

		if ((epgChannel.m_displayName == null) || epgChannel.m_displayName.isEmpty())
			epgChannel.m_displayName = id;

		channels.put(id, epgChannel);
		// DVScheduler.log("Adding Channel " + id + ":" + name);
	}

	private class Desc
	{
		int m_languagePriority;
		String m_languageCode, m_text;
	}

	public boolean parseProgram(Node prog, HashMap< String, ChannelEpgScan > channels, String name, SortedMap<String, GuideChannel> programs, Map<String, Integer> languages, int maxLanguage)
	{
		GuideItem item = new GuideItem();

		String temp01 = "";
		String temp02 = "";

		try
		{
			NamedNodeMap attMap = prog.getAttributes();

			Node start_time = attMap.getNamedItem("start");
			temp01 = start_time.getNodeValue();
			Date start = parseDate(start_time.getNodeValue());
			item.setStart(start);

			Node stop_time = attMap.getNamedItem("stop");
			temp02 = stop_time.getNodeValue();
			Date stop = parseDate(stop_time.getNodeValue());

			// if(stop.before(start))
			// stop.add(Calendar.HOUR, 24);

			long duration = (stop.getTime() - start.getTime());
			duration = duration / (1000 * 60);

			if(duration < 0)
				duration *= -1;
			else if (duration == 0)
				return false;

			item.setDuration((int) duration);

			Node chan = attMap.getNamedItem("channel");

			if (prog.hasChildNodes())
			{
				Node found = null;
				NodeList nl2 = prog.getChildNodes();
				List<Desc> titles = new LinkedList<Desc>(); 
				List<Desc> subTitles = new LinkedList<Desc>();
				List<Desc> descs = new LinkedList<Desc>();
				for (int x = 0; x < nl2.getLength(); x++)
				{
					found = nl2.item(x);
					if (found.getFirstChild() != null)
					{
						if (found.getNodeName().equals("title"))
						{
							Node nlang = found.getAttributes().getNamedItem("lang");
							String lang = (nlang == null) ? "" : nlang.getTextContent().toLowerCase();
							Desc desc = new Desc();
							Integer priority = languages.get(lang);
							desc.m_languagePriority = (priority == null) ? Integer.MAX_VALUE : priority;
							desc.m_languageCode = lang;
							desc.m_text = found.getFirstChild().getNodeValue();
							titles.add(desc);
						}
						else if (found.getNodeName().equals("sub-title"))
						{
							Node nlang = found.getAttributes().getNamedItem("lang");
							String lang = (nlang == null) ? "" : nlang.getTextContent().toLowerCase();
							Desc desc = new Desc();
							Integer priority = languages.get(lang);
							desc.m_languagePriority = (priority == null) ? Integer.MAX_VALUE : priority;
							desc.m_languageCode = lang;
							desc.m_text = found.getFirstChild().getNodeValue();
							subTitles.add(desc);
						}
						else if (found.getNodeName().equals("desc"))
						{
							Node nlang = found.getAttributes().getNamedItem("lang");
							String lang = (nlang == null) ? "" : nlang.getTextContent().toLowerCase();
							Desc desc = new Desc();
							Integer priority = languages.get(lang);
							desc.m_languagePriority = (priority == null) ? Integer.MAX_VALUE : priority;
							desc.m_languageCode = lang;
							desc.m_text = found.getFirstChild().getNodeValue();
							descs.add(desc);
							//desc2 += desc2.isEmpty() ? found.getFirstChild().getNodeValue() : "<br>" + found.getFirstChild().getNodeValue();
							//else if (found.getNodeName().equals("url"))
							//	item.setURL(found.getFirstChild().getNodeValue());
						}
						else if (found.getNodeName().equals("category"))
						{
							item.addCategory( found.getFirstChild().getNodeValue() );
						}
					}
				}
				Comparator<Desc> comparator = new Comparator<Desc>() {
					public int compare(Desc g1, Desc g2) {
						return g1.m_languagePriority - g2.m_languagePriority;
					}
				};
				Collections.sort(titles, comparator);
				Iterator<Desc> iter = titles.iterator();
				int count = 0;
				String title2 = "";
				while (iter.hasNext())
				{
					Desc desc = iter.next();
					//title2 += title2.isEmpty() ? desc.m_text : "<br>" + desc.m_text;
					title2 += title2.isEmpty() ? desc.m_text : " | " + desc.m_text;
					if (++count >= maxLanguage)
						break;
				}
				item.setName(title2);

				Collections.sort(subTitles, comparator);
				Iterator<Desc> iter2 = subTitles.iterator();
				count = 0;
				String subTitle2 = "";
				while (iter2.hasNext())
				{
					Desc desc = iter2.next();
					//subTitle2 += subTitle2.isEmpty() ? desc.m_text : "<br>" + desc.m_text;
					subTitle2 += subTitle2.isEmpty() ? desc.m_text : " | " + desc.m_text;
					if (++count >= maxLanguage)
						break;
				}
				item.setSubName(subTitle2);

				Collections.sort(descs, comparator);
				Iterator<Desc> iter3 = descs.iterator();
				count = 0;
				String desc2 = "";
				while (iter3.hasNext())
				{
					Desc desc = iter3.next();
					//desc2 += desc2.isEmpty() ? desc.m_text : "<br>" + desc.m_text;
					desc2 += desc2.isEmpty() ? desc.m_text : " | " + desc.m_text;
					if (++count >= maxLanguage)
						break;
				}
				item.setDescription(desc2);
			}

			Node url = getChildByType(prog, "url");
			if(url != null && url.getFirstChild() != null)
				item.setURL(url.getFirstChild().getNodeValue());

			// New Guide Info - XMLTV1.1 / XMLTV2 Loader Code..

			if(getNodeExists(prog, "rating"))
			{
				Node tmp = getChildByType(prog, "rating");
				Node value = getChildByType(tmp, "value");
				Node ratingNode = value.getFirstChild();
				if(ratingNode != null)
					item.setRatings(ratingNode.getNodeValue());
			}

			item.setLanguage(getNodeString(prog, "language", item.getLanguage()));

			item.setLastChance(getNodeExists(prog, "last-chance"));
			item.setRepeat(getNodeExists(prog, "previously-shown"));
			item.setPremiere(getNodeExists(prog, "premiere"));
			item.setCaptions(getNodeExists(prog, "subtitles"));

			if(getNodeExists(prog, "audio"))
			{
				Node tmp = getChildByType(prog, "audio");
				String str = getNodeString(tmp, "stereo", "").toLowerCase();
				if(str.startsWith("surround"))
					item.setSurround(true);
				if(str.startsWith("ac3"))
					item.setAC3(true);
				tmp = null;
				str = "";
			}

			if(getNodeExists(prog, "video"))
			{
				Node tmp = getChildByType(prog, "video");
				String str = getNodeString(tmp, "aspect", "");
				if(str.toLowerCase().startsWith("16:9"))
					item.setWidescreen(true);
				str = getNodeString(tmp, "quality", "");
				if(str.toLowerCase().startsWith("hd"))
					item.setHighDef(true);
				tmp = null;
				str = "";
			}

			String epgChannelId = chan.getNodeValue();
			ChannelEpgScan epgChannel = channels.get(epgChannelId);
			String epgChannelId2 = getEpgChannelId(name, epgChannel, epgChannel.m_originalNetworkId, epgChannel.m_transportStreamId, epgChannel.m_sourceId, epgChannel.m_programId, epgChannelId);

			/*
			channel = channel.replaceAll("'", "`");
			channel = channel.replaceAll("\"", "`");
			channel = channel.replaceAll("<", "(");
			channel = channel.replaceAll(">", ")");
			 */

			GuideChannel guideChannel = programs.get(epgChannelId2);
			if(guideChannel == null)
			{
				guideChannel = new GuideChannel(epgChannelId2, epgChannel.m_displayName, name, epgChannel.m_originalNetworkId, 
						epgChannel.m_transportStreamId, epgChannel.m_sourceId, epgChannel.m_programId);
				programs.put(epgChannelId2, guideChannel);
			}
			guideChannel.add(item);

			if (epgChannel.m_programId2 >= 0)
			{
				epgChannelId2 = getEpgChannelId(name, epgChannel, epgChannel.m_originalNetworkId2, epgChannel.m_transportStreamId2, epgChannel.m_sourceId, epgChannel.m_programId2, epgChannelId);
				programs.put(epgChannelId2, guideChannel);
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
			DVScheduler.log("Temp 01:" + temp01);
			DVScheduler.log("Temp 02:" + temp02);
			DVScheduler.log("Available Item Information:");
			DVScheduler.log("Name:" + item.getName());
			DVScheduler.log("Sub Name:" + item.getSubName());
			DVScheduler.log("StartTime:" + item.getStart());
			DVScheduler.log("Duration:" + item.getDuration());
			DVScheduler.log("Description:" + item.getDescription());

			return false;
		}

		return true;
	}

	public Node getChildByType(Node node, String name)
	{
		if(!node.hasChildNodes())
			return null;

		Node found = null;
		NodeList nl = node.getChildNodes();
		for (int x = 0; x < nl.getLength(); x++)
		{
			found = nl.item(x);
			if(found.getNodeName().equals(name))
				return found;
		}

		return null;
	}

	String[] getFirstLetters()
	{
		SortedSet<String> firsts = new TreeSet<String>();

		for (int x = 0; x < m_channelMaps.size(); x++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(x);
			for (int j=0; j<mapGroup.m_channelMaps.size(); j++)
			{
				ChannelMap map = mapGroup.m_channelMaps.get(j);
				GuideChannel guideChannel = m_programs.get(map.m_epgChannelId);

				if(guideChannel != null)
				{
					guideChannel.getFirstLetters(firsts);
				}
			}
		}

		String[] result = (String[]) firsts.toArray(new String[0]);
		return result;
	}

	String[] getNamesStartingWith(String starting)
	{
		SortedSet<String> starts = new TreeSet<String>();

		for (int x = 0; x < m_channelMaps.size(); x++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(x);
			for (int j=0; j<mapGroup.m_channelMaps.size(); j++)
			{
				ChannelMap map = mapGroup.m_channelMaps.get(j);
				GuideChannel guideChannel = m_programs.get(map.m_epgChannelId);

				if(guideChannel != null)
				{
					guideChannel.getNamesStartingWith(starting, starts);
				}
			}
		}

		String[] result = (String[]) starts.toArray(new String[0]);
		return result;
	}

	GuideItem[] getItems(String name, String channel)
	{
		GuideChannel progs = (GuideChannel) m_programs.get(channel);
		if(progs != null)
		{
			return progs.getItems(name);
		}
		else
			return new GuideItem[0];
	}

	public int searchEPG(EpgMatchList epgMatchList, HashMap<String, Vector<GuideItem>> result)
	{
		DataStore store = DataStore.getInstance();

		int num = 0;

		GuideItem[] progs = null;

		Vector<?> channelMap = getOldChannelMap();

		if (channelMap == null || channelMap.size() == 0)
		{
			return 0;
		}

		Set<?> wsChannels = store.getChannels().keySet();
		String[] titleFilters = (store.m_EpgTitleFilter == null || store.m_EpgTitleFilter.isEmpty()) ? null : DVScheduler.getArray(store.m_EpgTitleFilter);

		for(int x = 0; x < channelMap.size(); x++)
		{
			String[] map = (String[])channelMap.get(x);

			if(wsChannels.contains(map[0]))
			{
				progs = getProgramsForChannel(map[1]);

				if(progs.length > 0)
				{
					Vector<GuideItem> results = null;

					if(result.containsKey(map[0]))
					{
						results = (Vector<GuideItem>) result.get(map[0]);
					}
					else
					{
						results = new Vector<GuideItem>();
						result.put(map[0], results);
					}

					for (int y = 0; y < progs.length; y++)
					{
						if(epgMatchList.isMatch(progs[y], map[0], titleFilters))
						{
							if(!results.contains(progs[y]))
							{
								results.add(progs[y]);
								num++;
							}
						}
					}
				}
			}
		}

		return num;
	}   

	public int addEpgMatches(StringBuffer buff, int format, boolean bMarkOnly) throws Exception
	{
		clearAutoAddMatched();
		DataStore store = DataStore.getInstance();

		int num = 0;
		SimpleDateFormat df = new SimpleDateFormat("EE d HH:mm");

		boolean overlap = true;
		try
		{
			overlap = "1".equals(store.getProperty("Schedule.overlap", "1").trim());
		}
		catch (Exception e){}

		String type = "Capture.deftype";
		/*
      String type = "";
      try
      {
    	  type = store.getProperty("Capture.deftype", "");
      }
      catch (Exception e){}
		 */

		// removes all conflicts from list
		m_conflictList.clear();

		CaptureDeviceList devList = CaptureDeviceList.getInstance();

		//
		// now cycle through all the EPG Auto-Add items
		//
		EpgMatch epgMatch = null;
		Calendar cal = Calendar.getInstance();

		Vector<String[]> channelMap = new Vector<String[]>();
		for(int i=0; i<m_channelMaps.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMaps.get(i);
			ChannelMap map = getBestChannelMap(mapGroup);
			if (map == null)
				continue;
			String[] oldMapping = { mapGroup.m_channelName, map.m_epgChannelId };
			channelMap.add(oldMapping);
		}
		Set<String> wsChannels = store.getChannels().keySet();

		//
		// Check to see if we have any mapped channels at all, if not just return
		//
		if (channelMap == null || channelMap.size() == 0)
		{
			if (buff != null)
			{
				if (format == 1)
					buff.append("Channel map not set so could not run Auto-Add system.<br>\n");
				else
					buff.append("Channel map not set so could not run Auto-Add system.\n");

				DVScheduler.log("Channel map not set so could not run Auto-Add system.");
				return 0;
			}
		}

		if (buff != null)
		{
			if (format == 1)
				buff.append("About to run Auto-Add system.<br><br>\n");
			else
				buff.append("About to run Auto-Add system.\n\n");
		}

		GuideItem[] progs = null;
		GuideItem guidItem = null;
		ScheduleItem schItem = null;
		ScheduleItem prevScheduleItem = null;
		int conflictNum = 0;
		Set<GuideItem> addedItems = new HashSet<GuideItem>();
		Set<String> consideredItems = new HashSet<String>();

		Vector<EpgMatch> epgMatchList = store.getEpgAutoAdds();
		String[] titleFilters = (store.m_EpgTitleFilter == null || store.m_EpgTitleFilter.isEmpty()) ? null : DVScheduler.getArray(store.m_EpgTitleFilter);

		List<ScheduleItem> scheduleItems = new LinkedList<ScheduleItem>();
		DVScheduler.log("");
		//
		// For each Auto-Add item check the EPG programs for a match
		//      
		for (int x = 0; x < epgMatchList.size(); x++)
		{
			epgMatch = (EpgMatch) epgMatchList.get(x);

			if ( ((epgMatch.isEnabled() || epgMatch.isEnabledMark()) && !bMarkOnly) || 
					(epgMatch.isEnabledMark() && bMarkOnly) )
			{

				//
				// for each Auto-Add item search the EPG programs for matches
				//
				for (int z = 0; z < channelMap.size(); z++)
				{
					String[] chanMap = (String[])channelMap.get(z);

					if(wsChannels.contains(chanMap[0]))
					{
						progs = getProgramsForChannel(chanMap[1]);

						prevScheduleItem = null;
						for (int y = 0; y < progs.length; y++)
						{
							guidItem = progs[y];

							//
							// Ok so we have a guide item, lets see if it matches the
							// Auto-Add item
							//

							boolean b_FinalResult = false;
							Vector<String> matchNames = epgMatch.getMatchListNames();
							HashMap<String, EpgMatchList> matchLists = store.getEpgMatches();

							EpgMatchList matcher = null;

							for(int nameIndex = 0; nameIndex < matchNames.size(); nameIndex++)
							{
								String matchListName = (String)matchNames.get(nameIndex);
								matcher = (EpgMatchList)matchLists.get(matchListName);
								if(matcher != null)
								{
									if(matcher.isMatch(guidItem, chanMap[0], titleFilters))
									{
										b_FinalResult = true;
										break;
									}
								}
							}

							boolean b_AlreadyDone = addedItems.contains(guidItem);

							if(b_FinalResult && b_AlreadyDone)
							{
								AddConflictDetails acd = new AddConflictDetails(guidItem.getName());
								acd.setConflict("This EPG item has already been added<br>by an EPG Auto-Add item with higher priority.");
								acd.setReason("Already Matched");
								m_conflictList.add(acd);
							}

							else if(b_FinalResult && !b_AlreadyDone)
							{
								//
								// It looks like we have a match
								//

								//
								// Create a new Schedule Item from the data in the EPG
								// item
								//
								schItem = new ScheduleItem(guidItem, chanMap[0],
										type, store.m_rand.nextLong(), epgMatch.isAutoDelete());
								schItem.setEpgAutoAddName(epgMatch.getName2());
								schItem.setKeepFor(epgMatch.getKeepFor());
								schItem.setPostTask(epgMatch.getPostTask());
								schItem.setCapturePath(epgMatch.getCapturePath());
								schItem.setFilePattern(epgMatch.GetFileNamePattern());
								schItem.setCapType(epgMatch.getCaptureType());

								cal.setTime(schItem.getStart());
								cal.add(Calendar.MINUTE,
										(epgMatch.getStartBuffer() * -1));
								schItem.setDuration(guidItem.getDuration()
										+ guidItem.getDuration() * epgMatch.getEPGBuffer() / 60 
										+ epgMatch.getStartBuffer()
										+ epgMatch.getEndBuffer());
								schItem.setStart(cal);
								boolean bAddSchedule = true;

								if ( epgMatch.isMergeDoubleEpisode() && 
										(prevScheduleItem != null) && 
										//prevScheduleItem.getChannel().equals(schItem.getChannel()) && 
										schItem.isOverlapping(prevScheduleItem) )
								{
									bAddSchedule = false;
								}

								String itemDetails = schItem.getName()
										+ " ("
										+ df.format(schItem.getStart(),
												new StringBuffer(),
												new FieldPosition(0)) + " ("
												+ schItem.getDuration() + " min) "
												+ schItem.getChannel() + ") ";

								DVScheduler.log("Matched Item = " + schItem.getName() + " " + schItem.getStart() + " " + schItem);
								String guideStr = guidItem.getStart().getTime() + "," + guidItem.getStop().getTime() + "," + guidItem.getName().toLowerCase() + "," + guidItem.getSubName().toLowerCase();
								//
								// Now lets work out if we can actually add this
								// schedule item
								// to out schedule table without conflicts etc
								//
								String itemAction = "";
								StringBuffer conflictResult = new StringBuffer();
								String itemResult = "";

								boolean alreadyInList = isAlreadyInList(schItem);

								if(alreadyInList)
								{
									if (bMarkOnly)
									{
										guidItem.addFavoriteFlagger(epgMatch.getID());
										addedItems.add(guidItem);
										itemAction = "FAVORITE";
										itemResult = "Already in list";
										num++;
									}
									else
									{
										if (epgMatch.isEnabled())
										{
											itemAction = "ERROR";
											itemResult = "Already in list";
										}
										if (epgMatch.isEnabledMark())
										{
											itemAction += "(FAVORITE)";
											guidItem.addFavoriteFlagger(epgMatch.getID());
										}
									}
								}
								else if (isAlreadyInListEPG(schItem))
								{
									if (bMarkOnly)
									{
										guidItem.addFavoriteFlagger(epgMatch.getID());
										addedItems.add(guidItem);
										itemAction = "FAVORITE";
										itemResult = "Already in modified or merged list";
										num++;
									}
									else
									{
										if (epgMatch.isEnabled())
										{
											if (bAddSchedule)
											{
												prevScheduleItem = schItem;
											}
											itemAction = "ERROR";
											itemResult = "Already in modified or merged list";
										}
										if (epgMatch.isEnabledMark())
										{
											itemAction += "(FAVORITE)";
											guidItem.addFavoriteFlagger(epgMatch.getID());
										}
									}
								}
								else if ( isInEpgExclusion(guidItem, chanMap[0]) )
								{
									itemAction = "ERROR";
									itemResult = "Item in exclusion";
								}
								else if(cal.before(Calendar.getInstance()))
								{
									if (bMarkOnly)
									{
										guidItem.addFavoriteFlagger(epgMatch.getID());
										addedItems.add(guidItem);
										itemAction = "FAVORITE";
										itemResult = "Before now";
										num++;
									}
									else
									{
										if (epgMatch.isEnabled())
										{
											itemAction = "ERROR";
											itemResult = "Before now";
										}
										if (epgMatch.isEnabledMark())
										{
											itemAction += "(FAVORITE)";
											guidItem.addFavoriteFlagger(epgMatch.getID());
										}
									}
								}
								else if (epgMatch.isDuplicateDetection() && consideredItems.contains(guideStr))
								{
									if (bMarkOnly)
									{
										guidItem.addFavoriteFlagger(epgMatch.getID());
										addedItems.add(guidItem);
										itemAction = "FAVORITE";
										itemResult = "Duplicate show";
										num++;
									}
									else
									{
										if (epgMatch.isEnabled())
										{
											itemAction = "ERROR";
											itemResult = "Duplicate show";
										}
										if (epgMatch.isEnabledMark())
										{
											itemAction += "(FAVORITE)";
											guidItem.addFavoriteFlagger(epgMatch.getID());
										}
									}
								}
								else
								{
									Channel channel = store.getChannel(schItem.getChannel());
									if ( overlap || (channel == null) || channel.isGroup() || 
											(numConflicts(schItem, conflictResult, null) < devList.getDeviceCountByChannel(chanMap[0])) )
									{
										if (bMarkOnly)
										{
											guidItem.addFavoriteFlagger(epgMatch.getID());
											addedItems.add(guidItem);
											itemAction = "FAVORITE";
											num++;												
										}
										else
										{
											if (epgMatch.isEnabled())
											{
												if (bAddSchedule)
												{
													prevScheduleItem = schItem;
													//store.getScheduleList().put(schItem.toString(), schItem);
													scheduleItems.add(schItem);
													itemAction = "ADDED";
												}
												else
												{
													int duration = (int)((guidItem.getStop().getTime() - prevScheduleItem.getStart().getTime()) / 1000 / 60) - epgMatch.getStartBuffer();
													prevScheduleItem.setDuration(duration
															+ duration * epgMatch.getEPGBuffer() / 60 
															+ epgMatch.getStartBuffer()
															+ epgMatch.getEndBuffer());
													prevScheduleItem.addChild(schItem);
													itemAction = "MERGED";
												}
												addedItems.add(guidItem);
												num++;
											}
											if (epgMatch.isEnabledMark())
											{
												itemAction += "(FAVORITE)";
												guidItem.addFavoriteFlagger(epgMatch.getID());
											}
										}
									}
									else
									{
										if (bMarkOnly)
										{
											guidItem.addFavoriteFlagger(epgMatch.getID());
											addedItems.add(guidItem);
											itemAction = "FAVORITE";
											itemResult = "Item overlapped";
											num++;
										}
										else
										{
											if (epgMatch.isEnabled())
											{
												itemAction = "ERROR";
												itemResult = "Item overlapped";

												AddConflictDetails acd = new AddConflictDetails(itemDetails);
												acd.setConflict(conflictResult.toString());
												acd.setReason(itemResult);
												m_conflictList.add(acd);

												conflictNum++;
											}
											if (epgMatch.isEnabledMark())
											{
												itemAction += "(FAVORITE)";
												guidItem.addFavoriteFlagger(epgMatch.getID());
											}
										}
									}
								}

								consideredItems.add(guideStr);

								if (buff != null)
								{
									if(format == 1)
									{
										if(itemAction.contains("ERROR"))
											buff.append("<b><font color='#FF0000'>"
													+ HTMLEncoder.encode(itemAction) + "</font></b> - "
													+ HTMLEncoder.encode(itemDetails) + " (" + HTMLEncoder.encode(itemResult)
													+ ") <a class='infoNav' href='/servlet/EpgDataRes?action=26&id=" + guidItem.toString() + "&channel=" + Formatter.encodeURL(chanMap[0], "UTF-8")
													+ "'>EPG</a><br>\n");
										else
											buff.append("<b><font color='#00FF00'>"
													+ HTMLEncoder.encode(itemAction) + "</font></b> - "
													+ HTMLEncoder.encode(itemDetails) + " (" + HTMLEncoder.encode(itemResult)
													+ ") <a class='infoNav' href='/servlet/EpgDataRes?action=26&id=" + guidItem.toString() + "&channel=" + Formatter.encodeURL(chanMap[0], "UTF-8")
													+ "'>EPG</a><br>\n");
									}
									else
										buff.append(itemAction + " - " + itemDetails + " ("
												+ itemResult + ")\n");
								}

							}
						} // end program loop
					} // end if channel name
				} // end channel loop
			} // end enabled if
			else
			{
				Vector<?> items = epgMatch.getMatchListNames();
				String names = "";
				for(int q = 0; q < items.size(); q++)
				{
					String name = (String)items.get(q);
					if(format == 1)
					{
						name = name.replaceAll("<", "&lt;");
						name = name.replaceAll(">", "&gt;");
					}
					if(q == items.size()-1)
					{
						names += name;
					}
					else
						names += name + ", ";
				}
				if(names.length() == 0)
					names = "No Match List Assigned";

				if (buff != null)
				{
					if(format == 1)
					{            
						buff.append("Auto-Add (" + names + ") is disabled.<br>\n");
					}
					else
					{
						buff.append("Auto-Add (" + names + ") is disabled.\n");
					}
				}
			}
		}// end Auto-Add loop

		ThreadLock.getInstance().getLock();
		try
		{
			Iterator<ScheduleItem> iter = scheduleItems.iterator();
			while (iter.hasNext())
			{
				ScheduleItem scheduleItem = iter.next();
				store.getScheduleList().put(scheduleItem.toString(), scheduleItem);
			}
		}
		finally
		{
			ThreadLock.getInstance().releaseLock();
		}

		if (buff != null)
		{
			if (format == 1)
				buff.append("<br>Matched and added " + num + " items from the EPG data.<br>\n");
			else
				buff.append("\nMatched and added " + num + " items from the EPG data.\n");
		}

		if (buff != null && conflictNum > 0 && format == 1)
		{
			buff.append("<br>There were " + conflictNum + " conflicts, check the <a href='/servlet/EpgAutoAddDataRes?action=25'>CONFLICT REPORT PAGE</a> for more info.<br>\n");
		}

		return num;
	}

	public boolean isAlreadyInList(ScheduleItem schItem)
	{
		DataStore store = DataStore.getInstance();

		ScheduleItem[] itemsArray = (ScheduleItem[]) store.getScheduleList().values().toArray(new ScheduleItem[0]);

		for (int x = 0; x < itemsArray.length; x++)
		{
			ScheduleItem item = itemsArray[x];

			if ( schItem.equals(item) && ((item.getState() == ScheduleItem.WAITING) || (item.getState() == ScheduleItem.RUNNING)) )
				return true;
		}

		return false;
	}

	public boolean isAlreadyInListEPG(ScheduleItem schItem)
	{
		DataStore store = DataStore.getInstance();

		ScheduleItem[] itemsArray = (ScheduleItem[]) store.getScheduleList().values().toArray(new ScheduleItem[0]);

		for (int x = 0; x < itemsArray.length; x++)
		{
			ScheduleItem item = itemsArray[x];

			if (item.getType() != ScheduleItem.EPG)
				continue;
			else if(schItem.equalsEPG(item))
				return true;
		}

		return false;
	}

	private int numConflicts(ScheduleItem schItem, StringBuffer buff, Vector<Long> ignore)
	{
		DataStore store = DataStore.getInstance();

		SimpleDateFormat df = new SimpleDateFormat("EE d HH:mm");
		ScheduleItem[] itemsArray = (ScheduleItem[]) store.getScheduleList().values().toArray(new ScheduleItem[0]);
		int number = 0;

		for (int x = 0; x < itemsArray.length; x++)
		{
			ScheduleItem item = itemsArray[x];
			if (item.isOverlapping(schItem))
			{
				number++;
				buff.append(number + " : " + item.getName() + " (");
				buff.append(df.format(item.getStart(), new StringBuffer(), new FieldPosition(0)));
				buff.append(" " + item.getDuration() + " ");
				buff.append(item.getChannel() + ")<br>\n");
			}
		}

		Vector<Integer> list = new Vector<Integer>();
		getConflictCount(schItem.getStart(), schItem.getStop(), store.getChannel(schItem.getChannel()), list, ignore);

		int maxConf = 0;
		for(int q = 0; q < list.size(); q++)
		{
			Integer count = (Integer)list.get(q);
			if(maxConf < count.intValue())
				maxConf = count.intValue();
		}

		return maxConf;
	}

	private void getConflictCount(Date start, Date end, Channel channel, Vector<Integer> list, Vector<Long> ignore)
	{
		Calendar startCal = Calendar.getInstance();
		startCal.setTime(start);
		Calendar endCal = Calendar.getInstance();
		endCal.setTime(end);

		while(startCal.before(endCal))
		{
			Calendar span = Calendar.getInstance();
			span.setTime(startCal.getTime());
			span.add(Calendar.MINUTE, 1);

			int maxConflict = getMaxConflictCount(startCal.getTime(), span.getTime(), channel, ignore);

			list.add(new Integer(maxConflict));

			startCal.add(Calendar.MINUTE, 1);
		}   
	}

	private int getMaxConflictCount(Date start, Date end, Channel channel, Vector<Long> ignore)
	{
		DataStore store = DataStore.getInstance();

		ScheduleItem[] itemsArray = (ScheduleItem[]) store.getScheduleList().values().toArray(new ScheduleItem[0]);

		int thisCount = 0;

		String netType = channel.getNetworkType();
		String setting = channel.getSetting();
		HashSet<String> conflictSettings = new HashSet<String>();

		for (int x = 0; x < itemsArray.length; x++)
		{
			ScheduleItem item = itemsArray[x];
			Channel channel2 = store.getChannel(item.getChannel());
			if (channel2 == null || channel2.isGroup() || !channel2.getNetworkType().equals(netType))
				continue;

			String setting2 = channel2.getSetting();

			boolean overlap = false;

			if(item.getStart().getTime() >= start.getTime() && item.getStart().getTime() < end.getTime() && !setting.equals(setting2) && !conflictSettings.contains(setting2))
				overlap = true;

			else if(item.getStop().getTime() > start.getTime() && item.getStop().getTime() <= end.getTime() && !setting.equals(setting2) && !conflictSettings.contains(setting2))
				overlap = true;

			else if(item.getStart().getTime() <= start.getTime() && item.getStop().getTime() >= end.getTime() && !setting.equals(setting2) && !conflictSettings.contains(setting2))
				overlap = true;

			if(overlap)
			{
				conflictSettings.add(setting2);

				if(ignore != null)
				{
					boolean doAdd = true;
					for(int q = 0; q < ignore.size(); q++)
					{
						Long ignoreID = (Long)ignore.get(q);
						if(ignoreID.longValue() == item.id)
						{
							doAdd = false;
						}
					}
					if(doAdd)
						thisCount++;
				}
				else
				{
					thisCount++;
				}
			}
		}
		return thisCount;
	}   

	public Vector<AddConflictDetails> getConflictList()
	{
		return m_conflictList;
	}

	public int simpleEpgSearch(String lookFor, int lookSite, String cat, String chan, HashMap<String, Vector<GuideItem>> result)
	{
		Set<String> cat2 = new HashSet<String>();
		Set<String> chan2 = new HashSet<String>();

		cat2.add(cat.toUpperCase());
		chan2.add(chan.toUpperCase());

		return simpleEpgSearch(lookFor, lookSite, cat2, chan2, 0xffffffff, result);
	}

	//	Function assumes cat and chan values are upper-cases (for case-insensitive comparison).
	//	lookSite:
	//	Bit 1 means search title.
	//	Bit 2 means search description.
	//	flags:
	//	Bit 1 means search in exclusion list.
	//	Bit 2 means search in not exclusion list.
	//	Bit 3 means search in auto-add matched list.
	//	Bit 4 means search in not auto-add matched list.
	public int simpleEpgSearch(String lookFor, int lookSite, Set<String> cat, Set<String> chan, int flags, HashMap<String, 
			Vector<GuideItem>> result)
	{
		int num = 0;

		GuideItem[] progs = null;
		GuideItem guideItem = null;

		DataStore store = DataStore.getInstance();

		Vector<?> channelMap = getOldChannelMap();
		Set<?> wsChannels = store.getChannels().keySet();

		if (channelMap == null || channelMap.size() == 0)
		{
			DVScheduler.log("Channel map not set so could not do search.");
			return 0;
		}

		// Some fixes.
		if ( ((flags & 0x01) == 0) && ((flags & 0x02) == 0) )
			flags |= 0x03;
		if ( ((flags & 0x04) == 0) && ((flags & 0x08) == 0) )
			flags |= 0x0c;

		for (int x = 0; x < channelMap.size(); x++)
		{
			String[] chanMap = (String[]) channelMap.get(x);

			if(wsChannels.contains(chanMap[0]))
			{
				progs = getProgramsForChannel(chanMap[1]);

				if(progs.length > 0)
				{
					Vector<GuideItem> results = new Vector<GuideItem>();

					for (int y = 0; y < progs.length; y++)
					{
						guideItem = progs[y];

						boolean matches = false;
						boolean catMatches = false;
						boolean chanMatches = false;
						boolean flagMatches = false;

						if(((lookSite & 0x01) > 0) && guideItem.getName().toUpperCase().indexOf(lookFor.toUpperCase()) > -1)
							matches = true;

						else if(((lookSite & 0x01) > 0) && guideItem.getSubName().toUpperCase().indexOf(lookFor.toUpperCase()) > -1)
							matches = true;

						else if(((lookSite & 0x02) > 0) && guideItem.getDescription().toUpperCase().indexOf(lookFor.toUpperCase()) > -1)
							matches = true;

						if(cat.isEmpty() || cat.contains(""))
							catMatches = true;
						else
						{
							Vector<String> categories = guideItem.getCategories();
							if (categories != null)
							{
								for (int i=0; i<categories.size(); i++)
								{
									if (cat.contains(categories.get(i).toUpperCase()))
									{
										catMatches = true;
										break;
									}
								}
							}
						}

						if(chan.isEmpty() || chan.contains(""))
							chanMatches = true;

						//else if(chanMap[0].toUpperCase().indexOf(chan.toUpperCase()) > -1)
						else if (chan.contains(chanMap[0].toUpperCase()))
							chanMatches = true;

						if (flags == 0xffffffff)
							flagMatches = true;
						else
						{
							boolean bResult2 = false;
							boolean bTest = isInEpgExclusion(guideItem, chanMap[0]);

							if ( ((flags & 0x01) > 0) && bTest )
								bResult2 = true;
							else if ( ((flags & 0x02) > 0) && !bTest )
								bResult2 = true;

							if (bResult2 == true)
							{
								bResult2 = false;
								bTest = guideItem.isAutoMark();

								if ( ((flags & 0x04) > 0) && bTest )
									bResult2 = true;
								else if ( ((flags & 0x08) > 0) && !bTest )
									bResult2 = true;
							}

							if (bResult2)
								flagMatches = true;
						}

						if(catMatches && matches && chanMatches && flagMatches)
						{
							results.add(guideItem);
							num++;
						}
					}

					if(results.size() > 0)
						result.put(chanMap[0], results);
				}
			}
		}

		return num;
	}

	private void clearAutoAddMatched()
	{
		Iterator<GuideChannel> iter = m_programs.values().iterator();
		while (iter.hasNext())
		{
			iter.next().clearAutoAddMatched();
		}
	}

	public long getEpgLoadTime()
	{
		return m_epgLoadTime;
	}

	public void getProgramNames(Collection<String> col)
	{
		Iterator<GuideChannel> iter = m_programs.values().iterator();
		while (iter.hasNext())
		{
			iter.next().getNames(col);
		}
	}

	private String getValue(String text)
	{
		return text.substring(text.indexOf("->") + 2);
	}

	public synchronized void scheduleInternalEpgScan()
	{
		try
		{
			if (m_internalEpgScheduleQueue.isEmpty())
				return;
			DataStore store = DataStore.getInstance();
			EpgLoadTime loadTime = null;
			long now = new Date().getTime();
			if (m_bReloadInternalEpg)
			{
				loadTime = m_internalEpgScheduleQueue.peek();
			}
			else
			{
				if (!store.getProperty("guide.scheduleInternalEnabled", "0").equals("1"))
					return;

				int epgLoadPeriod = store.getPropertyInt("guide.scheduleInternalPeriod", 24) * 3600000;
				loadTime = m_internalEpgScheduleQueue.peek();
				if ((loadTime.m_loadTime + epgLoadPeriod) > now)
					return;
			}

			Channel channel = store.getChannel(loadTime.m_channelName);
			if ((channel == null) || channel.isGroup() || (channel.getProgramID() < 0) || !channel.isEpgScan())
				return;

			CaptureDeviceList deviceList = CaptureDeviceList.getInstance();
			CaptureDevice device = deviceList.GetFirstFreeDevice(channel.getNetworkType());
			if (device == null)
				return;

			// Remove and add to make sure everyone gets a turn.
			m_internalEpgScheduleQueue.remove();
			loadTime.m_loadTime = now;
			m_internalEpgScheduleQueue.add(loadTime);
			saveInternalEpgSchedule();

			// Check capture type.
			String captureEngineName = null;
			String capType = "scan.ce";
			String capTypeFile = store.getProperty("Capture.engine.path", "util") + File.separator + capType;
			File dir = new File(capTypeFile);

			if (!dir.exists() || !dir.isFile())
			{
				DVScheduler.log("Capture type file (" + capType + ") doesn't exist");
				return;
			}
			else if (dir.length() > 4096)
			{
				DVScheduler.log("Capture type file (" + capType + ") > 4096 bytes - too big, check for error");
				return;
			}

			BufferedReader r;
			r = new BufferedReader(new FileReader(capTypeFile));
			String line = null;

			while ((line = r.readLine()) != null)
			{
				line = line.trim();

				if (line.startsWith("Internal EPG Scan Engine->"))
				{
					captureEngineName = getValue(line).trim();
				}
			}

			r.close();

			CaptureEngine captureEngine = new CaptureEngine();
			if ( !captureEngine.SetEngine(captureEngineName, store.getProperty("Capture.engine.path", "util") + File.separator + captureEngineName + ".ce") )
			{
				DVScheduler.log("EPG scan error reading capture engine file (" + captureEngineName + ".ce)");

				return;
			}
			if (captureEngine.getInitialEpgScanCommand().isEmpty())
				return;
			CaptureProcess captureProcess = new CaptureProcess(device);
			captureProcess.startDeviceAndEpgScan(channel, captureEngine);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	private String getEpgChannelId(String setting, int onid, int tsid, int sid, int prid)
	{
		if ((onid >= 0) && (tsid >= 0) && (prid >= 0))
			return setting + "+OI" + onid + "_TI" + tsid + "_PI" + prid;

		if ((tsid >= 0) && (sid >= 0))
			return setting + "+TI" + tsid + "_SI" + sid;

		return null;
	}

	private String getEpgChannelId(Channel channel)
	{
		int onid = channel.getOriginalNetworkId();
		int tsid = channel.getTransportStreamId();
		int sid = channel.getSourceId();
		int prid = channel.getProgramID();

		return getEpgChannelId(channel.getSettingForFilename(), onid, tsid, sid, prid);
	}

	private String getEpgChannelId(String setting, ChannelEpgScan epgChannel, int onid, int tsid, int sid, int prid, String defaultEpgChannelId)
	{
		String epgChannelId2;
		if (epgChannel == null)
		{
			epgChannelId2 = defaultEpgChannelId;
		}
		else if (setting.isEmpty())	// File source.
		{
			epgChannelId2 = epgChannel.m_displayName + "_" + defaultEpgChannelId;
		}
		else	// Internal EPG scan.
		{
			epgChannelId2 = getEpgChannelId(setting, onid, tsid, sid, prid);
			if (epgChannelId2 == null)	// This should not happen
			{
				DVScheduler.log("Bug in GuideStore.parseProgram(.), notify author");
				epgChannelId2 = setting + "_" + epgChannel.m_displayName + "_" + defaultEpgChannelId;
			}
		}
		return epgChannelId2;
	}

	private boolean isEpgChannelId(Channel channel)
	{
		int onid = channel.getOriginalNetworkId();
		int tsid = channel.getTransportStreamId();
		int sid = channel.getSourceId();
		int prid = channel.getProgramID();

		if ((onid >= 0) && (tsid >= 0) && (prid >= 0))
			return true;

		if ((tsid >= 0) && (sid >= 0))
			return true;

		return false;
	}

	public synchronized boolean checkReloadInternalEpg()
	{
		if (!m_bReloadInternalEpg)
			return false;

		if (m_internalEpgScheduleQueue.isEmpty())
		{
			m_bReloadInternalEpg = false;
			return false;
		}

		EpgLoadTime loadTime = m_internalEpgScheduleQueue.peek();
		if (loadTime.m_loadTime == 0)
		{
			scheduleInternalEpgScan();
			return true;
		}
		else
		{
			m_bReloadInternalEpg = false;
			return false;
		}
	}

	protected synchronized void setInternalEpgScheduleManualReload()
	{
		if (m_bReloadInternalEpg)
			return;

		DataStore store = DataStore.getInstance();
		if (!store.getProperty("guide.source.type", "Internal").contains("Internal"))
		{
			m_internalEpgScheduleQueue.clear();
			saveInternalEpgSchedule();
			return;
		}
		m_internalEpgScheduleQueue.clear();
		HashSet<String> settingsScheduled = new HashSet<String>();
		Iterator<Channel> iter = store.getChannels().values().iterator();
		while (iter.hasNext())
		{
			Channel ch = iter.next();
			if (ch.isGroup() || (ch.getProgramID() < 0) || !ch.isEpgScan())
				continue;

			String setting = ch.getSettingForFilename();
			if (settingsScheduled.contains(setting))
				continue;

			m_internalEpgScheduleQueue.add(new EpgLoadTime(ch.getName(), setting, 0));
			settingsScheduled.add(setting);
		}
		m_bReloadInternalEpg = true;
		//saveInternalEpgSchedule();
	}

	private synchronized void setInternalEpgScheduleLoadedGuide(String setting)
	{
		DataStore store = DataStore.getInstance();
		if ( !store.getProperty("guide.source.type", "Internal").contains("Internal") || !store.getProperty("guide.scheduleInternalEnabled", "0").equals("1") )
		{
			m_internalEpgScheduleQueue.clear();
			saveInternalEpgSchedule();
		}
		else
		{
			// Go through the queue
			long now = new Date().getTime();
			Iterator<EpgLoadTime> iter2 = m_internalEpgScheduleQueue.iterator();
			while (iter2.hasNext())
			{
				EpgLoadTime loadTime = iter2.next();
				if (loadTime.m_setting.equals(setting))
				{
					loadTime.m_loadTime = now;
					break;
				}
			}
			saveInternalEpgSchedule();
		}

		ThreadLock.getInstance().getLock();
		try
		{
			Date dateSchedule = new Date();
			int epgLoadPeriod = store.getPropertyInt("guide.scheduleInternalPeriod", 24) * 3600000;
			long scheduleTime = dateSchedule.getTime() + epgLoadPeriod;
			dateSchedule.setTime(scheduleTime);
			Calendar calendarSchedule = Calendar.getInstance();
			calendarSchedule.setTime(dateSchedule);
			List<ScheduleItem> newSchedules = new LinkedList<ScheduleItem>();
			Map<String, ScheduleItem> schedules = store.getScheduleList();
			Iterator<ScheduleItem> iter3 = schedules.values().iterator();
			while (iter3.hasNext())
			{
				ScheduleItem item = iter3.next();
				if ( (item.getType() != ScheduleItem.EPG_SCAN) || (item.getState() != ScheduleItem.WAITING) )
					continue;

				Channel ch = store.getChannel(item.getChannel());
				if ((ch == null) || ch.isGroup() || !ch.getSettingForFilename().equals(setting))
					continue;

				if (item.getStart().getTime() < scheduleTime)
				{
					item.setState(ScheduleItem.FINISHED);
					item.setStatus("Finished");
					item.log("Finished/skipped due to recent internal EPG scan");
					ScheduleItem nextInstance = item.createNextInstance(calendarSchedule, store.m_rand);
					newSchedules.add(nextInstance);
				}
			}
			if (!newSchedules.isEmpty())
			{
				Iterator<ScheduleItem> iter = newSchedules.iterator();
				while (iter.hasNext())
				{
					ScheduleItem scheduleItem = iter.next();
					store.addScheduleItem(scheduleItem);
				}
				store.saveSchedules(null);
			}
		} catch (Exception e)
		{
			e.printStackTrace();
		} finally
		{
			ThreadLock.getInstance().releaseLock();
		}
	}

	protected synchronized void setInternalEpgScheduleChannelChange()
	{
		DataStore store = DataStore.getInstance();
		if ( !store.getProperty("guide.source.type", "Internal").contains("Internal") || !store.getProperty("guide.scheduleInternalEnabled", "0").equals("1") )
		{
			m_internalEpgScheduleQueue.clear();
			saveInternalEpgSchedule();
			return;
		}

		long now = new Date().getTime();

		// Go through the queue
		HashSet<String> settingsScheduled = new HashSet<String>();
		Iterator<EpgLoadTime> iter2 = m_internalEpgScheduleQueue.iterator();
		while (iter2.hasNext())
		{
			EpgLoadTime loadTime = iter2.next();
			Channel ch = store.getChannel(loadTime.m_channelName);
			if ((ch == null) || ch.isGroup() || (ch.getProgramID() < 0) || !ch.isEpgScan())
				iter2.remove();
			else
				settingsScheduled.add(loadTime.m_setting);
		}

		// Go through the channels
		Iterator<Channel> iter = store.getChannels().values().iterator();
		while (iter.hasNext())
		{
			Channel ch = iter.next();
			if (ch.isGroup() || (ch.getProgramID() < 0) || !ch.isEpgScan())
				continue;

			String setting = ch.getSettingForFilename();
			if (settingsScheduled.contains(setting))
				continue;

			m_internalEpgScheduleQueue.add(new EpgLoadTime(ch.getName(), setting, now));
			settingsScheduled.add(setting);
		}
		saveInternalEpgSchedule();
	}

	private synchronized void deleteOldEpg()
	{
		if (System.nanoTime() < (m_deletedOldEpgTime + 12L * 3600L * 1000000000L))
		{
			return;
		}

		DVScheduler.log("Deleting old EPG");
		DataStore store = DataStore.getInstance();
		long minTime = new Date().getTime() - store.getPropertyInt("EPG.oldestItem", 72) * 3600 * 1000;
		Iterator<Entry<String, GuideChannel>> iter2 = m_programs.entrySet().iterator();
		while (iter2.hasNext())
		{
			Entry<String, GuideChannel> entry = iter2.next();
			GuideChannel guideChannel = entry.getValue();
			guideChannel.removeOldItems(minTime);
			if (guideChannel.isEmpty())
			{
				DVScheduler.log("Deleting empty EPG: " + entry.getKey());
				iter2.remove();
			}
		}

		calculateDateRange();
		m_deletedOldEpgTime = System.nanoTime();
	}

	private boolean setChannel(Channel channel)
	{
		Iterator<GuideChannel> iter = m_programs.values().iterator();
		while (iter.hasNext())
		{
			GuideChannel guideChannel = iter.next();
			String setting = channel.getSettingForFilename();
			if (setting.equals(guideChannel.getChannelSetting()) && (channel.getProgramID() == guideChannel.getProgramId()))
			{
				int onid = channel.getOriginalNetworkId();
				int tsid = channel.getTransportStreamId();
				int sid = channel.getSourceId();

				int onid2 = guideChannel.getOriginalNetworkId();
				int tsid2 = guideChannel.getTransportStreamId();
				int sid2 = guideChannel.getSourceId();

				boolean bChanged = false;
				if ((onid2 >= 0) && (onid != onid2))
				{
					channel.setOriginalNetworkId(onid2);
					bChanged = true;
				}
				if ((tsid2 >= 0) && (tsid != tsid2))
				{
					channel.setTransportStreamId(tsid2);
					bChanged = true;
				}
				if ((sid2 >= 0) && (sid != sid2))
				{
					channel.setSourceId(sid2);
					bChanged = true;
				}
				return bChanged;
			}
		}
		return false;
	}

	private void setChannels()
	{
		DataStore store = DataStore.getInstance();

		boolean bSaveChannels = false;
		Iterator<Channel> iter = store.getChannels().values().iterator();
		while (iter.hasNext())
		{
			Channel ch = iter.next();
			if (ch.isGroup() || (ch.getProgramID() < 0))
				continue;

			String epgChannelId = getEpgChannelId(ch);
			if (epgChannelId != null)
				continue;

			if (setChannel(ch))
				bSaveChannels = true;
		}
		if (bSaveChannels)
		{
			try {
				store.saveChannels(null);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public void setCustomEpgChannelMapping()
	{
		m_channelMaps = m_channelMapsCustom;
	}

	public synchronized void setInternalEpgChannelMapping()
	{
		DataStore store = DataStore.getInstance();
		// Have to set default to custom due to version 5.72 and earlier not having default.
		if (store.getProperty("guide.mapping", "custom").equals("custom"))
		{
			m_channelMaps = m_channelMapsCustom;
			return;
		}
		else
			m_channelMaps = m_channelMapsInternal;

		m_channelMapsInternal.clear();
		Vector<String> channels = store.m_channelOrder.getChannels();
		for (int i=0; i<channels.size(); i++)
		{
			Channel channel = store.getChannel(channels.get(i));
			if ((channel == null) || channel.isGroup() || (channel.getProgramID() < 0))
				continue;

			/*
			String epgChannelId = getEpgChannelId(channel);
			if (epgChannelId == null)
				continue;

			GuideChannel guideChannel = m_programs.get(epgChannelId);
			if ((guideChannel == null) || guideChannel.isEmpty())
				continue;

			ChannelMapGroup mapGroup = new ChannelMapGroup();
			mapGroup.m_channelName = channels.get(i);

			ChannelMap map = new ChannelMap();
			map.m_epgChannelId = epgChannelId;
			map.m_epgChannelName = guideChannel.getEpgChannelName();
			map.m_setting = guideChannel.getChannelSetting();

			mapGroup.m_channelMaps.add(map);
			m_channelMapsInternal.add(mapGroup);
			 */

			if (!isEpgChannelId(channel))
				continue;

			ChannelMapGroup mapGroup = new ChannelMapGroup();
			mapGroup.m_channelName = channels.get(i);

			Iterator<Entry<String, GuideChannel>> iter = m_programs.entrySet().iterator();
			while (iter.hasNext())
			{
				Entry<String, GuideChannel> entry = iter.next();
				GuideChannel guideChannel = entry.getValue();
				if (!guideChannel.isEpgForChannel(channel))
					continue;

				ChannelMap map = new ChannelMap();
				map.m_epgChannelId = entry.getKey();
				map.m_epgChannelName = guideChannel.getEpgChannelName();
				map.m_setting = guideChannel.getChannelSetting();

				mapGroup.m_channelMaps.add(map);
			}
			if (!mapGroup.m_channelMaps.isEmpty())
				m_channelMapsInternal.add(mapGroup);
		}
	}

	public void addDefaultEpgMapping()
	{
		DataStore store = DataStore.getInstance();
		Vector<String> channels = store.m_channelOrder.getChannels();
		for (int i=0; i<channels.size(); i++)
		{
			Channel channel = store.getChannel(channels.get(i));
			if ((channel == null) || channel.isGroup() || (channel.getProgramID() < 0))
				continue;

			/*
			GuideChannel guideChannel = null;
			String epgChannelId = getEpgChannelId(channel);
			if (epgChannelId == null)
				continue;

			addChannelToCustomMap(channels.get(i), epgChannelId, 10);
			 */

			Iterator<Entry<String, GuideChannel>> iter = m_programs.entrySet().iterator();
			while (iter.hasNext())
			{
				Entry<String, GuideChannel> entry = iter.next();
				GuideChannel guideChannel = entry.getValue();
				if (!guideChannel.isEpgForChannel(channel))
					continue;

				addChannelToCustomMap(channels.get(i), entry.getKey(), -1);
			}

		}
		saveChannelMap();
	}

	public void loadInternalXmltv(File file)
	{
		deleteOldEpg();
		GuideDataLoader loader = new GuideDataLoader();
		byte[] data = null;
		String name = file.getName();
		name = name.substring(0, name.lastIndexOf('.'));
		try
		{
			data = loader.getFileContents(file.getAbsolutePath());
			file.delete();
			DataStore store = DataStore.getInstance();
			boolean bKeepOldData = (store.getPropertyInt("EPG.oldestItem", 72) > 0);
			int found = loadXmlData(data, name, bKeepOldData);
			DVScheduler.log("Loaded " + name + ": " + found + " events");
			if (found <= 0)
				return;

			setChannels();
			setInternalEpgChannelMapping();

			for (int i=0; i<m_channelMaps.size(); i++)
			{
				ChannelMapGroup mapGroup = m_channelMaps.get(i);
				if (mapGroup.isExistChannelSetting(name))
				{
					StringBuffer actionBuff = new StringBuffer();
					int autoMark = store.getProperty("guide.postInternalScanAutoMarkEnabled", "1").equals("0") ? 0 : 2;
					int autoAdd = store.getProperty("guide.postInternalScanAutoAddEnabled", "1").equals("0") ? 0 : 2;
					runPostLoadActions(actionBuff, 0, 1, autoMark, autoAdd);

					DVScheduler.log(actionBuff);
					break;
				}
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}

		m_epgLoadTime = new Date().getTime();

		saveEpg();
		setAndSaveCategories();
		setInternalEpgScheduleLoadedGuide(name);

		calculateDateRange();

		System.gc();

		DataStore store = DataStore.getInstance();
		String postTaskName = store.getProperty("Tasks.PostInternalScanTask", "");
		HashMap<String, TaskCommand> tasks = store.getTaskList();
		TaskCommand taskCommand = (TaskCommand) tasks.get(postTaskName);

		if(taskCommand != null)
		{
			StringBuffer buff = new StringBuffer();
			//runTask(taskCommand, true, buff, 0);
			runTask(taskCommand, false, buff, 0);
		}

		adjustSchedulesToEpg();
	}

	public void getChannelLastLoadTimeHtml(StringBuffer buff)
	{
		DateFormatter2 dateFormatter2 = new DateFormatter2();
		Date date = new Date();
		Iterator<Entry<String, GuideChannel>> iter = m_programs.entrySet().iterator();
		while (iter.hasNext())
		{
			Entry<String, GuideChannel> entry = iter.next();
			GuideChannel guideChannel = entry.getValue();
			buff.append("<tr>\n");
			buff.append("<td>" + HTMLEncoder.encode(guideChannel.getEpgChannelName() + " (" + entry.getKey()) + ")</td>\n");
			date.setTime(guideChannel.getLoadTime());
			buff.append("<td>" + dateFormatter2.formatDateTime(date) + "</td>\n");
			buff.append("</tr>\n");
		}
	}

	public void getInternalEpgScheduleHtml(StringBuffer buff)
	{
		if (m_internalEpgScheduleQueue.isEmpty())
			return;
		DataStore store = DataStore.getInstance();
		int epgLoadPeriod = store.getPropertyInt("guide.scheduleInternalPeriod", 24) * 3600000;
		Date date = new Date();
		DateFormatter2 dateFormatter2 = new DateFormatter2();
		Vector<EpgLoadTime> loadTimes = new Vector<EpgLoadTime>();
		loadTimes.addAll(m_internalEpgScheduleQueue);
		Collections.sort(loadTimes);
		for (int i=0; i<loadTimes.size(); i++)
		{
			EpgLoadTime loadTime = loadTimes.get(i);
			Channel channel = store.getChannel(loadTime.m_channelName);
			if (channel == null)
				continue;
			buff.append("<tr>\n");
			//String demod = channel.getDemodulatorName().isEmpty() ? channel.getSetting() : channel.getDemodulatorName();
			String demod = channel.getSetting();
			buff.append("<td>" + HTMLEncoder.encode(demod) + "</td>\n");
			date.setTime(loadTime.m_loadTime + epgLoadPeriod);
			buff.append("<td>" + dateFormatter2.formatDateTime(date) + "</td>\n");
			buff.append("</tr>\n");
		}
	}

	public boolean isChannelExist(String channelId)
	{
		return m_programs.containsKey(channelId);
	}

	public String getEpgChannelName(String epgChannelId)
	{
		GuideChannel guideChannel = m_programs.get(epgChannelId);
		return (guideChannel == null) ? null : guideChannel.getEpgChannelName();
	}

	public void sortChannelMapCustom()
	{
		DataStore store = DataStore.getInstance();
		for (int i=0; i<m_channelMapsCustom.size(); i++)
		{
			ChannelMapGroup mapGroup = m_channelMapsCustom.get(i);
			int index = store.m_channelOrder.getChannelNumber(mapGroup.m_channelName);
			if (index < 0)
				mapGroup.m_sort = Integer.MAX_VALUE;
			else
				mapGroup.m_sort = index;
		}

		Comparator<ChannelMapGroup> comparator = new Comparator<ChannelMapGroup>() {
			public int compare(ChannelMapGroup g1, ChannelMapGroup g2) {
				return g1.m_sort - g2.m_sort;
			}
		};

		Collections.sort(m_channelMapsCustom, comparator);
	}
}
