/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.text.SimpleDateFormat;
import java.util.*;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.w3c.dom.Document;

import java.util.zip.GZIPOutputStream;

public class XSL
{
	private Document m_document = null;
	private HTTPurl m_urlData = null;
	private String m_xslName = "";
	private Map<String, String> m_headers = null;
	private SimpleDateFormat df = new SimpleDateFormat("EEE, d MMM yyyy H:mm:ss");
	private Vector<String> cookies = new Vector<String>();

	protected class TemplateCache
	{
		Templates m_template;
		long m_time;
	}

	static private Map<String, TemplateCache> m_caches = new HashMap<String, TemplateCache>();

	public XSL(Document doc, String name, Map<String, String> headers, HTTPurl urlData)
	{
		m_document = doc;
		m_xslName = name;
		m_urlData = urlData;
		m_headers = headers;
	}

	public XSL(Document doc, String name, Map<String, String> headers, HTTPurl urlData, boolean bBackURL)
	{
		m_document = doc;
		m_xslName = name;
		m_urlData = urlData;
		m_headers = headers;
		if (bBackURL)
		{
			try
			{
				addCookie("backURL", m_urlData.getRequestString());
			}
			catch (Exception e) {}
		}
	}

	public void addCookie(String n, String d) throws Exception
	{
		String name = Formatter.encodeURL(n, "UTF-8");
		String data = Formatter.encodeURL(d, "UTF-8");
		cookies.add(name + "=" + data + "; PATH=/");
	}

	public byte[] transform() throws Exception
	{
		return transform(-1, false);
	}

	public byte[] transform(int nCompress, boolean bConvertHtmlBreak) throws Exception
	{
		DataStore store = DataStore.getInstance();

		String httpDir = store.getProperty("path.httproot", "http");
		String themeDir = store.getThemeDirectory(m_headers.get("User-Agent"));
		String path = httpDir + File.separator + "themes" + File.separator + themeDir + File.separator + "xsl" + File.separator + m_xslName;

		//
		// Send output back
		//
		ByteArrayOutputStream buff = new ByteArrayOutputStream();
		TransformerFactory factory = TransformerFactory.newInstance();
		Transformer xformer = null;

		if ( (nCompress < 0) || (nCompress > 1) )
		{
			String compress = store.getProperty("HTTP.compress", "Yes");
			if (compress.equals("Yes"))
				nCompress = 1;
			else
				nCompress = 0;
		}
		if (nCompress == 1)
		{
			if (m_headers != null)
			{
				String acceptEncoding = (String)m_headers.get("Accept-Encoding");
				if ( (acceptEncoding == null) || !acceptEncoding.toLowerCase().contains("gzip") || HTTPResponse.isLocal(m_headers) )
					nCompress = 0;
			}
			else
				nCompress = 0;
		}

		// Apply the XSL template or NOT
		if (!"1".equalsIgnoreCase(m_urlData.getParameter("xml")))
		{
			File fileIn = new File(path);
			if (fileIn.exists() == false)
			{
				path = httpDir + File.separator + "xsl" + File.separator + m_xslName;
				fileIn = new File(path);
				if(fileIn.exists() == false)
					throw new Exception("XSL file (" + path + ") not found");
			}

			TemplateCache cache = m_caches.get(fileIn.getAbsolutePath());
			if ((cache == null) || (fileIn.lastModified() != cache.m_time))
			{
				cache = new TemplateCache();
				cache.m_time = fileIn.lastModified();
				FileInputStream fis = new FileInputStream(fileIn);
				try
				{
					cache.m_template = factory.newTemplates(new StreamSource(fis));
				}
				catch (Exception e)
				{
					fis.close();
					throw new Exception("Error in " + path + "\n" + e.toString());
				}
				m_caches.put(fileIn.getAbsolutePath(), cache);

				xformer = cache.m_template.newTransformer();
			}
			else
			{
				xformer = cache.m_template.newTransformer();
			}

			if (nCompress == 1)
				buff.write("HTTP/1.0 200 OK\nContent-Encoding: gzip\nContent-Type: text/html\n".getBytes());
			else
				buff.write("HTTP/1.0 200 OK\nContent-Type: text/html\n".getBytes());
		}
		else
		{
			if (nCompress == 1)
				buff.write("HTTP/1.0 200 OK\nContent-Encoding: gzip\nContent-Type: text/xml\n".getBytes());
			else
				buff.write("HTTP/1.0 200 OK\nContent-Type: text/xml\n".getBytes());
			xformer = factory.newTransformer();
		}

		String timeStamp = df.format(new Date());
		timeStamp += " GMT";

		buff.write(("Pragma: no-cache\n").getBytes());
		buff.write(("Cache-Control: no-cache\n").getBytes());
		buff.write(("Last-Modified: " + timeStamp + "\n").getBytes());
		buff.write(("Date: " + timeStamp + "\n").getBytes());
		buff.write(("Expires: " + timeStamp + "\n").getBytes()); 

		for(int x = 0; x < cookies.size(); x++)
		{
			String cook = (String)cookies.get(x);
			buff.write(("Set-Cookie: " + cook + "\n").getBytes());
		}

		buff.write("\n".getBytes());

		Source source = new DOMSource(m_document);

		if (nCompress == 1)
		{
			if ( !"1".equalsIgnoreCase(m_urlData.getParameter("xml")) && bConvertHtmlBreak )
			{
				ByteArrayOutputStream out2 = new ByteArrayOutputStream();
				Result result = new StreamResult(out2);
				xformer.transform(source, result);

				GZIPOutputStream gzipOut = new GZIPOutputStream(buff);
				gzipOut.write(out2.toString("UTF-8").replace("&lt;br&gt;", "<br>").getBytes("UTF-8"));
				gzipOut.finish();
				gzipOut.flush();
				return buff.toByteArray();
			}
			else
			{
				GZIPOutputStream gzipOut = new GZIPOutputStream(buff);
				Result result = new StreamResult(gzipOut);
				xformer.transform(source, result);
				gzipOut.finish();
				gzipOut.flush();
				return buff.toByteArray();
			}
		}
		else
		{
			Result result = new StreamResult(buff);
			xformer.transform(source, result);

			if ( !"1".equalsIgnoreCase(m_urlData.getParameter("xml")) && bConvertHtmlBreak )
			{
				return buff.toString("UTF-8").replace("&lt;br&gt;", "<br>").getBytes("UTF-8");
			}
			else
			{
				return buff.toByteArray();
			}
		}
	}
}